import copy
import itertools as it
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Set, TYPE_CHECKING, Tuple, Union

import numpy as np
import torch
from torch import nn

from .helper_funcs import (
    _get_call_stack_dicts,
    get_attr_values_from_tensor_list,
    get_tensor_memory_amount,
    get_vars_of_type_from_obj,
    human_readable_size,
    index_nested,
    log_current_rng_states,
    make_random_barcode,
    make_short_barcode_from_input,
    make_var_iterable,
    safe_copy,
    tensor_nanequal,
)
from .flops import compute_backward_flops, compute_forward_flops
from .tensor_log import TensorLogEntry

if TYPE_CHECKING:
    from .model_history import ModelHistory
    from .tensor_log import TensorLogEntry


def save_new_activations(
    self: "ModelHistory",
    model: nn.Module,
    input_args: Union[torch.Tensor, List[Any]],
    input_kwargs: Dict[Any, Any] = None,
    layers_to_save: Union[str, List] = "all",
    random_seed: Optional[int] = None,
):
    """Saves activations to a new input to the model, replacing existing saved activations.
    This will be much faster than the initial call to log_forward_pass (since all the of the metadata has
    already been saved), so if you wish to save the activations to many different inputs for a given model
    this is the function you should use. The one caveat is that this function assumes that the computational
    graph will be the same for the new input; if the model involves a dynamic computational graph that can change
    across inputs, and this graph changes for the new input, then this function will throw an error. In that case,
    you'll have to do a new call to log_forward_pass to log the new graph.

    Args:
        model: Model for which to save activations
        input_args: Either a single tensor input to the model, or list of input arguments.
        input_kwargs: Dict of keyword arguments to the model.
        layers_to_save: List of layers to save, using any valid lookup keys
        random_seed: Which random seed to use
    Returns:
        Nothing, but now the ModelHistory object will have saved activations for the new input.
    """
    self.logging_mode = "fast"

    # Go through and clear all existing activations.
    for tensor_log_entry in self:
        tensor_log_entry.tensor_contents = None
        tensor_log_entry.has_saved_activations = False
        tensor_log_entry.has_saved_grad = False
        tensor_log_entry.grad_contents = None
        tensor_log_entry.has_child_tensor_variations = False
        tensor_log_entry.children_tensor_versions = {}

    # Reset relevant fields.
    self.layers_with_saved_activations = []
    self.layers_with_saved_gradients = []
    self.num_tensors_saved = 0
    self.tensor_fsize_saved = 0
    self._tensor_counter = 0
    self._raw_layer_type_counter = defaultdict(lambda: 0)

    # Now run and log the new inputs.
    self._run_and_log_inputs_through_model(
        model, input_args, input_kwargs, layers_to_save, random_seed
    )


def log_source_tensor(self, t: torch.Tensor, source: str, extra_address: Optional[str] = None):
    if self.logging_mode == "exhaustive":
        log_source_tensor_exhaustive(self, t, source, extra_address)
    elif self.logging_mode == "fast":
        log_source_tensor_fast(self, t, source)


def log_source_tensor_exhaustive(
    self, t: torch.Tensor, source: str, extra_addr: Optional[str] = None
):
    """Takes in an input or buffer tensor, marks it in-place with relevant information, and
    adds it to the log.

    Args:
        t: the tensor
        source: either 'input' or 'buffer'
        extra_addr: either the buffer address or the input address
    """
    layer_type = source
    # Fetch counters and increment to be ready for next tensor to be logged
    self._tensor_counter += 1
    self._raw_layer_type_counter[layer_type] += 1
    realtime_tensor_num = self._tensor_counter
    layer_type_num = self._raw_layer_type_counter[layer_type]

    tensor_label = f"{layer_type}_{layer_type_num}_raw"

    if source == "input":
        is_input_layer = True
        has_input_ancestor = True
        input_output_address = extra_addr
        is_buffer_layer = False
        buffer_address = None
        buffer_parent = None
        initialized_inside_model = False
        has_internally_initialized_ancestor = False
        input_ancestors = {tensor_label}
        internally_initialized_ancestors = set()
        operation_equivalence_type = (
            f"input_{'_'.join(tuple(str(s) for s in t.shape))}_{str(t.dtype)}"
        )
    elif source == "buffer":
        is_input_layer = False
        has_input_ancestor = False
        input_output_address = None
        is_buffer_layer = True
        buffer_address = extra_addr
        initialized_inside_model = True
        has_internally_initialized_ancestor = True
        internally_initialized_ancestors = {tensor_label}
        input_ancestors = set()
        operation_equivalence_type = f"buffer_{extra_addr}"
        if hasattr(t, "tl_buffer_parent"):
            buffer_parent = t.tl_buffer_parent
        else:
            buffer_parent = None
    else:
        raise ValueError("source must be either 'input' or 'buffer'")

    fields_dict = {
        # General info:
        "tensor_label_raw": tensor_label,
        "layer_label_raw": tensor_label,
        "realtime_tensor_num": realtime_tensor_num,
        "operation_num": None,
        "source_model_history": self,
        "_pass_finished": False,
        # Label Info:
        "layer_label": None,
        "layer_label_short": None,
        "layer_label_w_pass": None,
        "layer_label_w_pass_short": None,
        "layer_label_no_pass": None,
        "layer_label_no_pass_short": None,
        "layer_type": layer_type,
        "layer_type_num": layer_type_num,
        "layer_total_num": None,
        "pass_num": 1,
        "layer_passes_total": 1,
        "lookup_keys": [],
        # Saved tensor info:
        "tensor_contents": None,
        "has_saved_activations": False,
        "activation_postfunc": self.activation_postfunc,
        "detach_saved_tensor": self.detach_saved_tensors,
        "output_device": self.output_device,
        "function_args_saved": False,
        "creation_args": None,
        "creation_kwargs": None,
        "tensor_shape": tuple(t.shape),
        "tensor_dtype": t.dtype,
        "tensor_fsize": get_tensor_memory_amount(t),
        "tensor_fsize_nice": human_readable_size(get_tensor_memory_amount(t)),
        # Child tensor variation tracking
        "has_child_tensor_variations": False,
        "children_tensor_versions": {},
        # Grad info:
        "grad_contents": None,
        "save_gradients": self.save_gradients,
        "has_saved_grad": False,
        "grad_shape": None,
        "grad_dtype": None,
        "grad_fsize": 0,
        "grad_fsize_nice": human_readable_size(0),
        # Function call info:
        "func_applied": None,
        "func_applied_name": "none",
        "func_call_stack": _get_call_stack_dicts(),
        "func_time_elapsed": 0,
        "flops_forward": 0,
        "flops_backward": 0,
        "func_rng_states": log_current_rng_states(),
        "func_argnames": tuple([]),
        "num_func_args_total": 0,
        "num_position_args": 0,
        "num_keyword_args": 0,
        "func_position_args_non_tensor": [],
        "func_keyword_args_non_tensor": {},
        "func_all_args_non_tensor": [],
        "function_is_inplace": False,
        "gradfunc": "none",
        "is_part_of_iterable_output": False,
        "iterable_output_index": None,
        # Param info:
        "computed_with_params": False,
        "parent_params": [],
        "parent_param_barcodes": [],
        "parent_param_passes": {},
        "num_param_tensors": 0,
        "parent_param_shapes": [],
        "num_params_total": int(0),
        "parent_params_fsize": 0,
        "parent_params_fsize_nice": human_readable_size(0),
        # Corresponding layer info:
        "operation_equivalence_type": operation_equivalence_type,
        "equivalent_operations": self.equivalent_operations[operation_equivalence_type],
        "same_layer_operations": [],
        # Graph info:
        "parent_layers": [],
        "has_parents": False,
        "parent_layer_arg_locs": {"args": {}, "kwargs": {}},
        "orig_ancestors": {tensor_label},
        "child_layers": [],
        "has_children": False,
        "sibling_layers": [],
        "has_siblings": False,
        "spouse_layers": [],
        "has_spouses": False,
        "is_input_layer": is_input_layer,
        "has_input_ancestor": has_input_ancestor,
        "input_ancestors": input_ancestors,
        "min_distance_from_input": None,
        "max_distance_from_input": None,
        "is_output_layer": False,
        "is_output_parent": False,
        "is_last_output_layer": False,
        "is_output_ancestor": False,
        "output_descendents": set(),
        "min_distance_from_output": None,
        "max_distance_from_output": None,
        "input_output_address": input_output_address,
        "is_buffer_layer": is_buffer_layer,
        "buffer_address": buffer_address,
        "buffer_pass": None,
        "buffer_parent": buffer_parent,
        "initialized_inside_model": initialized_inside_model,
        "has_internally_initialized_ancestor": has_internally_initialized_ancestor,
        "internally_initialized_parents": [],
        "internally_initialized_ancestors": internally_initialized_ancestors,
        "terminated_inside_model": False,
        # Conditional info:
        "is_terminal_bool_layer": False,
        "is_atomic_bool_layer": False,
        "atomic_bool_val": None,
        "in_cond_branch": False,
        "cond_branch_start_children": [],
        # Module info:
        "is_computed_inside_submodule": False,
        "containing_module_origin": None,
        "containing_modules_origin_nested": [],
        "module_nesting_depth": 0,
        "modules_entered": [],
        "modules_entered_argnames": defaultdict(list),
        "module_passes_entered": [],
        "is_submodule_input": False,
        "modules_exited": [],
        "module_passes_exited": [],
        "is_submodule_output": False,
        "is_bottom_level_submodule_output": False,
        "bottom_level_submodule_pass_exited": None,
        "module_entry_exit_threads_inputs": [],
        "module_entry_exit_thread_output": [],
    }

    _make_tensor_log_entry(self, t, fields_dict, (), {}, self.activation_postfunc)

    # Tag the tensor itself with its label, and with a reference to the model history log.
    t.tl_tensor_label_raw = tensor_label

    # Log info to ModelHistory
    self.equivalent_operations[operation_equivalence_type].add(t.tl_tensor_label_raw)
    if source == "input":
        self.input_layers.append(tensor_label)
    if source == "buffer":
        self.buffer_layers.append(tensor_label)
        self.internally_initialized_layers.append(tensor_label)

    # Make it track gradients if relevant

    if self.save_gradients:
        _add_backward_hook(self, t, t.tl_tensor_label_raw)


def log_source_tensor_fast(self, t: torch.Tensor, source: str):
    """NOTES TO SELF--fields to change are:
    for ModelHistory: pass timing, num tensors saved, tensor fsize
    for Tensors: tensor contents, fsize, args, kwargs, make sure to clear gradients.
    Add a minimal postprocessing thing for tallying stuff pertaining to saved tensors.
    Have some minimal checker to make sure the graph didn't change.
    """
    layer_type = source
    # Fetch counters and increment to be ready for next tensor to be logged
    self._tensor_counter += 1
    self._raw_layer_type_counter[layer_type] += 1
    layer_type_num = self._raw_layer_type_counter[layer_type]

    tensor_label_raw = f"{layer_type}_{layer_type_num}_raw"
    t.tl_tensor_label_raw = tensor_label_raw
    if tensor_label_raw in self.orphan_layers:
        return
    orig_tensor_label = self._raw_to_final_layer_labels[tensor_label_raw]
    if orig_tensor_label in self.unlogged_layers:
        return
    orig_tensor_entry = self.layer_dict_all_keys[orig_tensor_label]
    if (self._tensor_nums_to_save == "all") or (
        orig_tensor_entry.realtime_tensor_num in self._tensor_nums_to_save
    ):
        self.layers_with_saved_activations.append(orig_tensor_entry.layer_label)
        orig_tensor_entry.save_tensor_data(
            t, [], {}, self.save_function_args, self.activation_postfunc
        )

    orig_tensor_entry.tensor_shape = tuple(t.shape)
    orig_tensor_entry.tensor_dtype = t.dtype
    orig_tensor_entry.tensor_fsize = get_tensor_memory_amount(t)
    orig_tensor_entry.tensor_fsize_nice = human_readable_size(get_tensor_memory_amount(t))


def log_function_output_tensors(
    self,
    func: Callable,
    func_name: str,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    arg_copies: Tuple[Any],
    kwarg_copies: Dict[str, Any],
    out_orig: Any,
    func_time_elapsed: float,
    func_rng_states: Dict,
    is_bottom_level_func: bool,
):
    if self.logging_mode == "exhaustive":
        log_function_output_tensors_exhaustive(
            self,
            func,
            func_name,
            args,
            kwargs,
            arg_copies,
            kwarg_copies,
            out_orig,
            func_time_elapsed,
            func_rng_states,
            is_bottom_level_func,
        )
    elif self.logging_mode == "fast":
        log_function_output_tensors_fast(
            self,
            func_name,
            args,
            kwargs,
            arg_copies,
            kwarg_copies,
            out_orig,
            func_time_elapsed,
            func_rng_states,
            is_bottom_level_func,
        )


def log_function_output_tensors_exhaustive(
    self,
    func: Callable,
    func_name: str,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    arg_copies: Tuple[Any],
    kwarg_copies: Dict[str, Any],
    out_orig: Any,
    func_time_elapsed: float,
    func_rng_states: Dict,
    is_bottom_level_func: bool,
):
    """Logs tensor or set of tensors that were computed from a function call.

    Args:
        func: function that was called
        args: positional arguments to function that was called
        kwargs: keyword arguments to function that was called
        arg_copies: copies of positional arguments to function that was called
        kwarg_copies: copies of keyword arguments to function that was called
        out_orig: original output from function
        func_time_elapsed: time it took for the function to run
        func_rng_states: states of the random number generator when the function is called
        is_bottom_level_func: whether the function is at the bottom-level of function nesting
    """
    # Unpacking and reformatting:
    layer_type = func_name.lower().replace("_", "")
    all_args = list(args) + list(kwargs.values())

    fields_dict = {}  # dict storing the arguments for initializing the new log entry

    non_tensor_args = [arg for arg in args if not _check_if_tensor_arg(arg)]
    non_tensor_kwargs = {key: val for key, val in kwargs.items() if not _check_if_tensor_arg(val)}
    arg_tensors = get_vars_of_type_from_obj(all_args, torch.Tensor, [torch.nn.Parameter])
    parent_layer_labels = get_attr_values_from_tensor_list(arg_tensors, "tl_tensor_label_raw")
    parent_layer_entries = [self[label] for label in parent_layer_labels]

    # General info
    fields_dict["layer_type"] = layer_type
    fields_dict["detach_saved_tensor"] = self.detach_saved_tensors
    fields_dict["output_device"] = self.output_device

    # Grad info:
    fields_dict["grad_contents"] = None
    fields_dict["save_gradients"] = self.save_gradients
    fields_dict["has_saved_grad"] = False
    fields_dict["grad_shape"] = None
    fields_dict["grad_dtype"] = None
    fields_dict["grad_fsize"] = 0
    fields_dict["grad_fsize_nice"] = human_readable_size(0)

    # Function call info
    fields_dict["func_applied"] = func
    fields_dict["func_applied_name"] = func_name
    fields_dict["func_call_stack"] = _get_call_stack_dicts()
    fields_dict["func_time_elapsed"] = func_time_elapsed
    fields_dict["func_rng_states"] = func_rng_states
    fields_dict["func_argnames"] = self.func_argnames[func_name.strip("_")]
    fields_dict["num_func_args_total"] = len(args) + len(kwargs)
    fields_dict["num_position_args"] = len(args)
    fields_dict["num_keyword_args"] = len(kwargs)
    fields_dict["func_position_args_non_tensor"] = non_tensor_args
    fields_dict["func_keyword_args_non_tensor"] = non_tensor_kwargs
    fields_dict["func_all_args_non_tensor"] = non_tensor_args + list(non_tensor_kwargs.values())

    # Graph info
    parent_layer_arg_locs = _get_parent_tensor_function_call_location(
        self, parent_layer_entries, args, kwargs
    )
    (
        input_ancestors,
        internally_initialized_ancestors,
    ) = _get_ancestors_from_parents(parent_layer_entries)
    internal_parent_layer_labels = [
        label for label in parent_layer_labels if self[label].has_internally_initialized_ancestor
    ]

    fields_dict["parent_layers"] = parent_layer_labels
    fields_dict["parent_layer_arg_locs"] = parent_layer_arg_locs
    fields_dict["has_parents"] = len(fields_dict["parent_layers"]) > 0
    fields_dict["orig_ancestors"] = input_ancestors.union(internally_initialized_ancestors)
    fields_dict["child_layers"] = []
    fields_dict["has_children"] = False
    fields_dict["sibling_layers"] = []
    fields_dict["has_siblings"] = False
    fields_dict["spouse_layers"] = []
    fields_dict["has_spouses"] = False
    fields_dict["is_input_layer"] = False
    fields_dict["has_input_ancestor"] = len(input_ancestors) > 0
    fields_dict["input_ancestors"] = input_ancestors
    fields_dict["min_distance_from_input"] = None
    fields_dict["max_distance_from_input"] = None
    fields_dict["is_output_layer"] = False
    fields_dict["is_output_parent"] = False
    fields_dict["is_last_output_layer"] = False
    fields_dict["is_output_ancestor"] = False
    fields_dict["output_descendents"] = set()
    fields_dict["min_distance_from_output"] = None
    fields_dict["max_distance_from_output"] = None
    fields_dict["input_output_address"] = None
    fields_dict["is_buffer_layer"] = False
    fields_dict["buffer_address"] = None
    fields_dict["buffer_pass"] = None
    fields_dict["buffer_parent"] = None
    fields_dict["initialized_inside_model"] = len(parent_layer_labels) == 0
    fields_dict["has_internally_initialized_ancestor"] = len(internally_initialized_ancestors) > 0
    fields_dict["internally_initialized_parents"] = internal_parent_layer_labels
    fields_dict["internally_initialized_ancestors"] = internally_initialized_ancestors
    fields_dict["terminated_inside_model"] = False
    fields_dict["is_terminal_bool_layer"] = False
    fields_dict["in_cond_branch"] = False
    fields_dict["cond_branch_start_children"] = []

    # Param info
    arg_parameters = get_vars_of_type_from_obj(all_args, torch.nn.parameter.Parameter)
    parent_param_passes = _process_parent_param_passes(arg_parameters)
    indiv_param_barcodes = list(parent_param_passes.keys())

    fields_dict["computed_with_params"] = len(parent_param_passes) > 0
    fields_dict["parent_params"] = arg_parameters
    fields_dict["parent_param_barcodes"] = indiv_param_barcodes
    fields_dict["parent_param_passes"] = parent_param_passes
    fields_dict["num_param_tensors"] = len(arg_parameters)
    fields_dict["parent_param_shapes"] = [tuple(param.shape) for param in arg_parameters]
    fields_dict["num_params_total"] = int(
        np.sum([np.prod(shape) for shape in fields_dict["parent_param_shapes"]])
    )
    fields_dict["parent_params_fsize"] = int(
        np.sum([get_tensor_memory_amount(p) for p in arg_parameters])
    )
    fields_dict["parent_params_fsize_nice"] = human_readable_size(
        fields_dict["parent_params_fsize"]
    )

    # Module info

    containing_modules_origin_nested = _get_input_module_info(self, arg_tensors)
    if len(containing_modules_origin_nested) > 0:
        is_computed_inside_submodule = True
        containing_module_origin = containing_modules_origin_nested[-1]
    else:
        is_computed_inside_submodule = False
        containing_module_origin = None

    fields_dict["is_computed_inside_submodule"] = is_computed_inside_submodule
    fields_dict["containing_module_origin"] = containing_module_origin
    fields_dict["containing_modules_origin_nested"] = containing_modules_origin_nested
    fields_dict["module_nesting_depth"] = len(containing_modules_origin_nested)
    fields_dict["modules_entered"] = []
    fields_dict["modules_entered_argnames"] = defaultdict(list)
    fields_dict["module_passes_entered"] = []
    fields_dict["is_submodule_input"] = False
    fields_dict["modules_exited"] = []
    fields_dict["module_passes_exited"] = []
    fields_dict["is_submodule_output"] = False
    fields_dict["is_bottom_level_submodule_output"] = False
    fields_dict["bottom_level_submodule_pass_exited"] = None
    fields_dict["module_entry_exit_threads_inputs"] = {
        p.tensor_label_raw: p.module_entry_exit_thread_output[:] for p in parent_layer_entries
    }
    fields_dict["module_entry_exit_thread_output"] = []

    is_part_of_iterable_output = any(
        [issubclass(type(out_orig), cls) for cls in [list, tuple, dict, set]]
    )
    fields_dict["is_part_of_iterable_output"] = is_part_of_iterable_output
    out_iter = make_var_iterable(out_orig)  # so we can iterate through it

    for i, out in enumerate(out_iter):
        if not _output_should_be_logged(out, is_bottom_level_func):
            continue

        fields_dict_onetensor = {key: copy.copy(value) for key, value in fields_dict.items()}
        fields_to_deepcopy = [
            "parent_layer_arg_locs",
            "containing_modules_origin_nested",
            "parent_param_passes",
        ]
        for field in fields_to_deepcopy:
            fields_dict_onetensor[field] = copy.deepcopy(fields_dict[field])
        _log_info_specific_to_single_function_output_tensor(
            self, out, i, args, kwargs, parent_param_passes, fields_dict_onetensor
        )
        _make_tensor_log_entry(
            self,
            out,
            fields_dict=fields_dict_onetensor,
            t_args=arg_copies,
            t_kwargs=kwarg_copies,
            activation_postfunc=self.activation_postfunc,
        )
        new_tensor_entry = self[fields_dict_onetensor["tensor_label_raw"]]
        new_tensor_label = new_tensor_entry.tensor_label_raw
        _update_tensor_family_links(self, new_tensor_entry)

        # Update relevant fields of ModelHistory
        # Add layer to relevant fields of ModelHistory:
        if fields_dict["initialized_inside_model"]:
            self.internally_initialized_layers.append(new_tensor_label)
        if fields_dict["has_input_ancestor"] and any(
            [
                (
                    self[parent_layer].has_internally_initialized_ancestor
                    and not self[parent_layer].has_input_ancestor
                )
                for parent_layer in fields_dict_onetensor["parent_layers"]
            ]
        ):
            self._layers_where_internal_branches_merge_with_input.append(new_tensor_label)

        # Tag the tensor itself with its label, and add a backward hook if saving gradients.
        out.tl_tensor_label_raw = fields_dict_onetensor["tensor_label_raw"]
        if self.save_gradients:
            _add_backward_hook(self, out, out.tl_tensor_label_raw)

        # Track if any parent's tensor_contents differs from what this child received.
        # This catches getitem slicing, view mutations, and any other case where
        # a parent's saved values don't match the arg copy the child actually saw.
        for parent_label in new_tensor_entry.parent_layers:
            parent = self[parent_label]
            if parent.has_saved_activations and self.save_function_args:
                parent_tensor_contents = _get_parent_contents(
                    parent_label,
                    arg_copies,
                    kwarg_copies,
                    new_tensor_entry.parent_layer_arg_locs,
                )
                if not tensor_nanequal(parent_tensor_contents, parent.tensor_contents):
                    parent.children_tensor_versions[new_tensor_entry.tensor_label_raw] = (
                        parent_tensor_contents
                    )
                    parent.has_child_tensor_variations = True


def _get_parent_contents(parent_label, arg_copies, kwarg_copies, parent_layer_arg_locs):
    """Utility function to get the value of a parent layer from the arguments passed to a function."""
    for pos, label in parent_layer_arg_locs["args"].items():
        if label == parent_label:
            return index_nested(arg_copies, pos)
    for argname, label in parent_layer_arg_locs["kwargs"].items():
        if label == parent_label:
            return index_nested(kwarg_copies, argname)
    raise ValueError("Parent layer not found in function arguments.")


def log_function_output_tensors_fast(
    self,
    func_name: str,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    arg_copies: Tuple[Any],
    kwarg_copies: Dict[str, Any],
    out_orig: Any,
    func_time_elapsed: float,
    func_rng_states: Dict,
    is_bottom_level_func: bool,
):
    # Collect information.
    layer_type = func_name.lower().replace("_", "")
    all_args = list(args) + list(kwargs.values())
    non_tensor_args = [arg for arg in args if not _check_if_tensor_arg(arg)]
    non_tensor_kwargs = {key: val for key, val in kwargs.items() if not _check_if_tensor_arg(val)}

    arg_tensors = get_vars_of_type_from_obj(all_args, torch.Tensor, [torch.nn.Parameter])
    out_iter = make_var_iterable(out_orig)

    for i, out in enumerate(out_iter):
        if not _output_should_be_logged(out, is_bottom_level_func):
            continue
        self._tensor_counter += 1
        self._raw_layer_type_counter[layer_type] += 1
        realtime_tensor_num = self._tensor_counter
        layer_type_num = self._raw_layer_type_counter[layer_type]
        tensor_label_raw = f"{layer_type}_{layer_type_num}_{realtime_tensor_num}_raw"
        if tensor_label_raw in self.orphan_layers:
            continue
        parent_layer_labels_raw = get_attr_values_from_tensor_list(
            arg_tensors, "tl_tensor_label_raw"
        )
        parent_layer_labels_orig = [
            self._raw_to_final_layer_labels[raw_label]
            for raw_label in parent_layer_labels_raw
            if raw_label in self._raw_to_final_layer_labels
        ]
        out.tl_tensor_label_raw = tensor_label_raw
        if tensor_label_raw not in self._raw_to_final_layer_labels:
            raise ValueError(
                "The computational graph changed for this forward pass compared to the original "
                "call to log_forward_pass (either due to different inputs or a different "
                "random seed), so save_new_activations failed. Please re-run "
                "log_forward_pass with the desired inputs."
            )
        orig_tensor_label = self._raw_to_final_layer_labels[tensor_label_raw]
        if orig_tensor_label in self.unlogged_layers:
            continue
        orig_tensor_entry = self[orig_tensor_label]

        if self.save_gradients:
            _add_backward_hook(self, out, orig_tensor_label)

        # Check to make sure the graph didn't change.
        if any(
            [
                orig_tensor_entry.realtime_tensor_num != self._tensor_counter,
                orig_tensor_entry.layer_type != layer_type,
                orig_tensor_entry.tensor_label_raw != tensor_label_raw,
                set(orig_tensor_entry.parent_layers) != set(parent_layer_labels_orig),
            ]
        ):
            raise ValueError(
                "The computational graph changed for this forward pass compared to the original "
                "call to log_forward_pass (either due to different inputs or a different "
                "random seed), so save_new_activations failed. Please re-run "
                "log_forward_pass with the desired inputs."
            )

        # Update any relevant fields.
        if (self._tensor_nums_to_save == "all") or (
            orig_tensor_entry.realtime_tensor_num in self._tensor_nums_to_save
        ):
            self.layers_with_saved_activations.append(orig_tensor_entry.layer_label)
            orig_tensor_entry.save_tensor_data(
                out,
                arg_copies,
                kwarg_copies,
                self.save_function_args,
                self.activation_postfunc,
            )
            for child_layer in orig_tensor_entry.child_layers:
                if child_layer in self.output_layers:
                    child_output = self.layer_dict_main_keys[child_layer]
                    if (
                        orig_tensor_entry.has_child_tensor_variations
                        and child_layer in orig_tensor_entry.children_tensor_versions
                    ):
                        # children_tensor_versions already has transforms applied
                        tensor_to_save = orig_tensor_entry.children_tensor_versions[child_layer]
                        child_output.tensor_contents = safe_copy(tensor_to_save)
                    else:
                        child_output.tensor_contents = safe_copy(out)
                        if self.activation_postfunc is not None:
                            child_output.tensor_contents = self.activation_postfunc(
                                child_output.tensor_contents
                            )
                    child_output.has_saved_activations = True
                    child_output.tensor_fsize = get_tensor_memory_amount(
                        child_output.tensor_contents
                    )
                    child_output.tensor_fsize_nice = human_readable_size(child_output.tensor_fsize)

        orig_tensor_entry.tensor_shape = tuple(out.shape)
        orig_tensor_entry.tensor_dtype = out.dtype
        orig_tensor_entry.tensor_fsize = get_tensor_memory_amount(out)
        orig_tensor_entry.tensor_fsize_nice = human_readable_size(get_tensor_memory_amount(out))
        orig_tensor_entry.func_time_elapsed = func_time_elapsed
        orig_tensor_entry.func_rng_states = func_rng_states
        orig_tensor_entry.func_position_args_non_tensor = non_tensor_args
        orig_tensor_entry.func_keyword_args_non_tensor = non_tensor_kwargs


def _output_should_be_logged(out: Any, is_bottom_level_func: bool) -> bool:
    """Function to check whether to log the output of a function.

    Returns:
        True if the output should be logged, False otherwise.
    """
    if type(out) is not torch.Tensor:  # only log if it's a tensor
        return False

    if (not hasattr(out, "tl_tensor_label_raw")) or is_bottom_level_func:
        return True
    else:
        return False


def _add_backward_hook(self, t: torch.Tensor, tensor_label):
    """Adds a backward hook to the tensor that saves the gradients to ModelHistory if specified.

    Args:
        t: tensor

    Returns:
        Nothing; it changes the tensor in place.
    """

    def log_grad_to_model_history(grad):
        _log_tensor_grad(self, grad, tensor_label)

    if (t.grad_fn is not None) or t.requires_grad:
        t.register_hook(log_grad_to_model_history)


def _log_info_specific_to_single_function_output_tensor(
    self,
    t: torch.Tensor,
    i: int,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    parent_param_passes: Dict[str, int],
    fields_dict: Dict[str, Any],
):
    """Function to log handle the logging of info that's specific to a single output tensor
    (e.g., the shape), and not common to all output tensors.

    Args:
        t: tensor to log
        i: index of the tensor in the function output
        args: positional args to the function that created the tensor
        kwargs: keyword args to the function that created the tensor
        parent_param_passes: Dict mapping barcodes of parent params to how many passes they've seen
        fields_dict: dictionary of fields with which to initialize the new TensorLogEntry
    """
    layer_type = fields_dict["layer_type"]
    indiv_param_barcodes = list(parent_param_passes.keys())
    self._tensor_counter += 1
    self._raw_layer_type_counter[layer_type] += 1
    realtime_tensor_num = self._tensor_counter
    layer_type_num = self._raw_layer_type_counter[layer_type]
    tensor_label_raw = f"{layer_type}_{layer_type_num}_{realtime_tensor_num}_raw"

    if len(parent_param_passes) > 0:
        operation_equivalence_type = _make_raw_param_group_barcode(indiv_param_barcodes, layer_type)
        fields_dict["operation_equivalence_type"] = operation_equivalence_type
        self.layers_computed_with_params[operation_equivalence_type].append(tensor_label_raw)
        fields_dict["pass_num"] = len(self.layers_computed_with_params[operation_equivalence_type])
    else:
        operation_equivalence_type = _get_operation_equivalence_type(
            args, kwargs, i, layer_type, fields_dict
        )
        fields_dict["operation_equivalence_type"] = operation_equivalence_type
        fields_dict["pass_num"] = 1

    self.equivalent_operations[operation_equivalence_type].add(tensor_label_raw)
    fields_dict["equivalent_operations"] = self.equivalent_operations[operation_equivalence_type]

    fields_dict["function_is_inplace"] = hasattr(t, "tl_tensor_label_raw")
    fields_dict["gradfunc"] = type(t.grad_fn).__name__

    if fields_dict["is_part_of_iterable_output"]:
        fields_dict["iterable_output_index"] = i
    else:
        fields_dict["iterable_output_index"] = None

    if (t.dtype == torch.bool) and (t.dim()) == 0:
        fields_dict["is_atomic_bool_layer"] = True
        fields_dict["atomic_bool_val"] = t.item()
    else:
        fields_dict["is_atomic_bool_layer"] = False
        fields_dict["atomic_bool_val"] = None

    # General info
    fields_dict["tensor_label_raw"] = tensor_label_raw
    fields_dict["layer_total_num"] = None
    fields_dict["same_layer_operations"] = []
    fields_dict["realtime_tensor_num"] = realtime_tensor_num
    fields_dict["operation_num"] = None
    fields_dict["source_model_history"] = self
    fields_dict["_pass_finished"] = False

    # Other labeling info
    fields_dict["layer_label"] = None
    fields_dict["layer_label_short"] = None
    fields_dict["layer_label_w_pass"] = None
    fields_dict["layer_label_w_pass_short"] = None
    fields_dict["layer_label_no_pass"] = None
    fields_dict["layer_label_no_pass_short"] = None
    fields_dict["layer_type"] = layer_type
    fields_dict["layer_label_raw"] = tensor_label_raw
    fields_dict["layer_type_num"] = layer_type_num
    fields_dict["pass_num"] = 1
    fields_dict["layer_passes_total"] = 1
    fields_dict["lookup_keys"] = []

    # Saved tensor info
    fields_dict["tensor_contents"] = None
    fields_dict["has_saved_activations"] = False
    fields_dict["activation_postfunc"] = self.activation_postfunc
    fields_dict["function_args_saved"] = False
    fields_dict["creation_args"] = None
    fields_dict["creation_kwargs"] = None
    fields_dict["tensor_shape"] = tuple(t.shape)
    fields_dict["tensor_dtype"] = t.dtype
    fields_dict["tensor_fsize"] = get_tensor_memory_amount(t)
    fields_dict["tensor_fsize_nice"] = human_readable_size(fields_dict["tensor_fsize"])

    # FLOPs computation
    fields_dict["flops_forward"] = compute_forward_flops(
        fields_dict.get("func_applied_name"),
        fields_dict["tensor_shape"],
        fields_dict.get("parent_param_shapes", []),
        args,
        kwargs,
    )
    fields_dict["flops_backward"] = compute_backward_flops(
        fields_dict.get("func_applied_name"),
        fields_dict["flops_forward"],
    )

    # Child tensor variation tracking
    fields_dict["has_child_tensor_variations"] = False
    fields_dict["children_tensor_versions"] = {}

    # If internally initialized, fix this information:
    if len(fields_dict["parent_layers"]) == 0:
        fields_dict["initialized_inside_model"] = True
        fields_dict["has_internally_initialized_ancestor"] = True
        fields_dict["internally_initialized_parents"] = []
        fields_dict["internally_initialized_ancestors"] = {tensor_label_raw}


def _make_tensor_log_entry(
    self,
    t: torch.Tensor,
    fields_dict: Dict,
    t_args: Optional[Tuple] = None,
    t_kwargs: Optional[Dict] = None,
    activation_postfunc: Optional[Callable] = None,
):
    """
    Given a tensor, adds it to the model_history, additionally saving the activations and input
    arguments if specified. Also tags the tensor itself with its raw tensor label
    and a pointer to ModelHistory.

    Args:
        t: tensor to log
        fields_dict: dictionary of fields to log in TensorLogEntry
        t_args: Positional arguments to the function that created the tensor
        t_kwargs: Keyword arguments to the function that created the tensor
        activation_postfunc: Function to apply to activations before saving them.
    """
    if t_args is None:
        t_args = []
    if t_kwargs is None:
        t_kwargs = {}

    new_entry = TensorLogEntry(fields_dict)
    if (self._tensor_nums_to_save == "all") or (
        new_entry.realtime_tensor_num in self._tensor_nums_to_save
    ):
        new_entry.save_tensor_data(
            t, t_args, t_kwargs, self.save_function_args, activation_postfunc
        )
        self.layers_with_saved_activations.append(new_entry.tensor_label_raw)
    self._raw_tensor_dict[new_entry.tensor_label_raw] = new_entry
    self._raw_tensor_labels_list.append(new_entry.tensor_label_raw)

    return new_entry


def _log_tensor_grad(self, grad: torch.Tensor, tensor_label_raw: str):
    """Logs the gradient for a tensor during a backward pass.

    Args:
        grad: the gradient
        tensor_label_raw: the raw tensor label

    Returns:

    """
    self.has_saved_gradients = True
    tensor_label = self._raw_to_final_layer_labels[tensor_label_raw]
    tensor_log_entry = self[tensor_label]
    layers_to_update = [tensor_label]
    if tensor_log_entry.is_output_parent:  # also update any linked outputs
        for child_layer in tensor_log_entry.child_layers:
            if self[child_layer].is_output_layer:
                layers_to_update.append(child_layer)

    for layer_label in layers_to_update:
        layer = self[layer_label]
        if layer_label not in self.layers_with_saved_gradients:
            self.layers_with_saved_gradients.append(layer_label)

        layer.log_tensor_grad(grad)

    layer_order = {layer: i for i, layer in enumerate(self.layer_labels)}
    self.layers_with_saved_gradients = sorted(
        self.layers_with_saved_gradients, key=lambda x: layer_order[x]
    )


def _check_if_tensor_arg(arg: Any) -> bool:
    """Helper function to check if an argument either is a tensor or is a list/tuple containing a tensor.

    Args:
        arg: argument

    Returns:
        True if arg is or contains a tensor, false otherwise
    """
    if issubclass(type(arg), torch.Tensor):
        return True
    elif type(arg) in [list, tuple]:
        for elt in arg:
            if issubclass(type(elt), torch.Tensor):
                return True
        return False
    elif type(arg) == dict:
        for val in arg.values():
            if issubclass(type(val), torch.Tensor):
                return True
        return False
    else:
        return False


def _get_parent_tensor_function_call_location(
    self,
    parent_log_entries: List[TensorLogEntry],
    args: Tuple[Any],
    kwargs: Dict[Any, Any],
) -> Dict:
    """Utility function that takes in the parent tensors, the args, and kwargs, and returns a dict specifying
    where in the function call the parent tensors were used.

    Args:
        parent_log_entries: List of parent tensors.
        args: Tuple of function args
        kwargs: Dict of function kwargs

    Returns:
        Dict that itself contains two dicts, one specifing which args are associated with parent tensors,
        and another specifing which kwargs are associated with parent tensors.
    """
    tensor_all_arg_positions = {"args": {}, "kwargs": {}}
    arg_struct_dict = {"args": args, "kwargs": kwargs}

    for parent_entry in parent_log_entries:
        for arg_type in ["args", "kwargs"]:
            arg_struct = arg_struct_dict[arg_type]
            _find_arg_positions_for_single_parent(
                parent_entry, arg_type, arg_struct, tensor_all_arg_positions
            )

    return tensor_all_arg_positions


def _find_arg_positions_for_single_parent(
    parent_entry: TensorLogEntry,
    arg_type: str,
    arg_struct: Union[List, Tuple, Dict],
    tensor_all_arg_positions: Dict,
):
    """Helper function that finds where a single parent tensor is used in either the args or kwargs of a function,
    and updates a dict that tracks this information.

    Args:
        parent_entry: Parent tensor
        arg_type: 'args' or 'kwargs'
        arg_struct: args or kwargs
        tensor_all_arg_positions: dict tracking where the tensors are used
    """
    iterfunc_dict = {
        "args": enumerate,
        "kwargs": lambda x: x.items(),
        list: enumerate,
        tuple: enumerate,
        dict: lambda x: x.items(),
    }
    iterfunc = iterfunc_dict[arg_type]

    for arg_key, arg in iterfunc(arg_struct):
        if getattr(arg, "tl_tensor_label_raw", -1) == parent_entry.tensor_label_raw:
            tensor_all_arg_positions[arg_type][arg_key] = parent_entry.tensor_label_raw
        elif type(arg) in [list, tuple, dict]:
            iterfunc2 = iterfunc_dict[type(arg)]
            for sub_arg_key, sub_arg in iterfunc2(arg):
                if getattr(sub_arg, "tl_tensor_label_raw", -1) == parent_entry.tensor_label_raw:
                    tensor_all_arg_positions[arg_type][(arg_key, sub_arg_key)] = (
                        parent_entry.tensor_label_raw
                    )


def _get_ancestors_from_parents(
    parent_entries: List[TensorLogEntry],
) -> Tuple[Set[str], Set[str]]:
    """Utility function to get the ancestors of a tensor based on those of its parent tensors.

    Args:
        parent_entries: list of parent entries

    Returns:
        List of input ancestors and internally initialized ancestors.
    """
    input_ancestors = set()
    internally_initialized_ancestors = set()

    for parent_entry in parent_entries:
        input_ancestors.update(parent_entry.input_ancestors)
        internally_initialized_ancestors.update(parent_entry.internally_initialized_ancestors)
    return input_ancestors, internally_initialized_ancestors


def _update_tensor_family_links(self, entry_to_update: TensorLogEntry):
    """For a given tensor, updates family information for its links to parents, children, siblings, and
    spouses, in both directions (i.e., mutually adding the labels for each family pair).

    Args:
        entry_to_update: dict of information about the TensorLogEntry to be created
    """
    tensor_label = entry_to_update.tensor_label_raw
    parent_tensor_labels = entry_to_update.parent_layers

    # Add the tensor as child to its parents

    for parent_tensor_label in parent_tensor_labels:
        parent_tensor = self[parent_tensor_label]
        if tensor_label not in parent_tensor.child_layers:
            parent_tensor.child_layers.append(tensor_label)
            parent_tensor.has_children = True

    # Set the parents of the tensor as spouses to each other

    for spouse1, spouse2 in it.combinations(parent_tensor_labels, 2):
        if spouse1 not in self[spouse2].spouse_layers:
            self[spouse2].spouse_layers.append(spouse1)
            self[spouse2].has_spouses = True
        if spouse2 not in self[spouse1].spouse_layers:
            self[spouse1].spouse_layers.append(spouse2)
            self[spouse1].has_spouses = True

    # Set the children of its parents as siblings to each other.

    for parent_tensor_label in parent_tensor_labels:
        _add_sibling_labels_for_new_tensor(self, entry_to_update, self[parent_tensor_label])


def _add_sibling_labels_for_new_tensor(
    self, entry_to_update: TensorLogEntry, parent_tensor: TensorLogEntry
):
    """Given a tensor and specified parent tensor, adds sibling labels to that tensor, and
    adds itself as a sibling to all existing children.

    Args:
        entry_to_update: the new tensor
        parent_tensor: the parent tensor
    """
    new_tensor_label = entry_to_update.tensor_label_raw
    for sibling_tensor_label in parent_tensor.child_layers:
        if sibling_tensor_label == new_tensor_label:
            continue
        sibling_tensor = self[sibling_tensor_label]
        sibling_tensor.sibling_layers.append(new_tensor_label)
        sibling_tensor.has_siblings = True
        entry_to_update.sibling_layers.append(sibling_tensor_label)
        entry_to_update.has_siblings = True


def _process_parent_param_passes(
    arg_parameters: List[torch.nn.Parameter],
) -> Dict[str, int]:
    """Utility function to mark the parameters with barcodes, and log which pass they're on.

    Args:
        arg_parameters: List of arg parameters

    Returns:

    """
    parent_param_passes = {}
    for param in arg_parameters:
        if not hasattr(param, "tl_param_barcode"):
            param_barcode = make_random_barcode()
            param.tl_param_barcode = param_barcode
            param.tl_pass_num = 1
        else:
            param_barcode = param.tl_param_barcode
            param.tl_pass_num += 1
        parent_param_passes[param_barcode] = param.tl_pass_num
    return parent_param_passes


def _make_raw_param_group_barcode(indiv_param_barcodes: List[str], layer_type: str):
    """Given list of param barcodes and layer type, returns the raw barcode for the
    param_group; e.g., conv2d_abcdef_uvwxyz

    Args:
        indiv_param_barcodes: List of barcodes for each individual parameter tensor
        layer_type: The layer type.

    Returns:
        Raw barcode for the param group
    """
    param_group_barcode = f"{layer_type}_{'_'.join(sorted(indiv_param_barcodes))}"
    return param_group_barcode


def _get_operation_equivalence_type(
    args: Tuple, kwargs: Dict, i: int, layer_type: str, fields_dict: Dict
):
    arg_hash = _get_hash_from_args(args, kwargs)
    operation_equivalence_type = f"{layer_type}_{arg_hash}"
    if fields_dict["is_part_of_iterable_output"]:
        operation_equivalence_type += f"_outindex{i}"
    if fields_dict["containing_module_origin"] is not None:
        module_str = fields_dict["containing_module_origin"][0]
        operation_equivalence_type += f"_module{module_str}"
    return operation_equivalence_type


def _get_hash_from_args(args, kwargs):
    """
    Get a hash from the args and kwargs of a function call, excluding any tracked tensors.
    Preserves positional arg indices, kwarg names, and dict keys to avoid collisions.
    """
    args_to_hash = []
    for a, arg in enumerate(args):
        _append_arg_hash(arg, f"pos{a}", args_to_hash)
    for key, arg in kwargs.items():
        _append_arg_hash(arg, f"kw_{key}", args_to_hash)

    if len(args_to_hash) == 0:
        return "no_args"
    return make_short_barcode_from_input(args_to_hash)


def _append_arg_hash(arg, prefix, args_to_hash):
    """Append hash-relevant info for a single arg, preserving structure."""
    if hasattr(arg, "tl_tensor_label_raw"):
        args_to_hash.append(f"{prefix}_tensor{arg.shape}")
    elif isinstance(arg, dict):
        for k, v in arg.items():
            _append_arg_hash(v, f"{prefix}_dk{k}", args_to_hash)
    elif isinstance(arg, (list, tuple, set)):
        for i, elem in enumerate(arg):
            _append_arg_hash(elem, f"{prefix}_i{i}", args_to_hash)
    elif isinstance(arg, torch.nn.Parameter):
        pass  # exclude parameters from hash (same as before)
    else:
        args_to_hash.append(f"{prefix}_{arg}")


def _update_tensor_containing_modules(tensor_entry: TensorLogEntry) -> List[str]:
    """Utility function that updates the containing modules of a Tensor by starting from the containing modules
    as of the last function call, then looks at the sequence of module transitions (in or out of a module) as of
    the last module it saw, and updates accordingly.

    Args:
        tensor_entry: Log entry of tensor to check

    Returns:
        List of the updated containing modules.
    """
    containing_modules = tensor_entry.containing_modules_origin_nested[:]
    thread_modules = tensor_entry.module_entry_exit_thread_output[:]
    for thread_module in thread_modules:
        if thread_module[0] == "+":
            containing_modules.append(thread_module[1:])
        elif (thread_module[0] == "-") and (thread_module[1:] in containing_modules):
            containing_modules.remove(thread_module[1:])
    return containing_modules


def _get_input_module_info(self, arg_tensors: List[torch.Tensor]):
    """Utility function to extract information about module entry/exit from input tensors.

    Args:
        arg_tensors: List of input tensors

    Returns:
        Variables with module entry/exit information
    """
    max_input_module_nesting = 0
    most_nested_containing_modules = []
    for t in arg_tensors:
        tensor_label = getattr(t, "tl_tensor_label_raw", -1)
        if tensor_label == -1:
            continue
        tensor_entry = self[tensor_label]
        containing_modules = _update_tensor_containing_modules(tensor_entry)
        if len(containing_modules) > max_input_module_nesting:
            max_input_module_nesting = len(containing_modules)
            most_nested_containing_modules = containing_modules[:]
    return most_nested_containing_modules



import copy
import os
import random
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd
import torch
from torch import nn
from tqdm import tqdm

from .helper_funcs import get_vars_of_type_from_obj, set_random_seed, warn_parallel
from .model_history import (
    ModelHistory,
)


def run_model_and_save_specified_activations(
    model: nn.Module,
    input_args: Union[torch.Tensor, List[Any]],
    input_kwargs: Dict[Any, Any],
    layers_to_save: Optional[Union[str, List[Union[int, str]]]] = "all",
    keep_unsaved_layers: bool = True,
    output_device: str = "same",
    activation_postfunc: Optional[Callable] = None,
    mark_input_output_distances: bool = False,
    detach_saved_tensors: bool = False,
    save_function_args: bool = False,
    save_gradients: bool = False,
    random_seed: Optional[int] = None,
) -> ModelHistory:
    """Internal function that runs the given input through the given model, and saves the
    specified activations, as given by the tensor numbers (these will not be visible to the user;
    they will be generated from the nicer human-readable names and then fed in).

    Args:
        model: PyTorch model.
        input_args: Input arguments to the model's forward pass: either a single tensor, or a list of arguments.
        input_kwargs: Keyword arguments to the model's forward pass.
        layers_to_save: List of layers to save
        keep_unsaved_layers: Whether to keep layers in the ModelHistory log if they don't have saved activations.
        output_device: device where saved tensors will be stored: either 'same' to keep unchanged, or
            'cpu' or 'cuda' to move to cpu or cuda.
        activation_postfunc: Function to apply to activations before saving them (e.g., any averaging)
        mark_input_output_distances: Whether to compute the distance of each layer from the input or output.
            This is computationally expensive for large networks, so it is off by default.
        detach_saved_tensors: whether to detach the saved tensors, so they remain attached to the computational graph
        save_function_args: whether to save the arguments to each function
        save_gradients: whether to save gradients from any subsequent backward pass
        random_seed: Which random seed to use.

    Returns:
        ModelHistory object with full log of the forward pass
    """
    model_name = str(type(model).__name__)
    model_history = ModelHistory(
        model_name,
        output_device,
        activation_postfunc,
        keep_unsaved_layers,
        save_function_args,
        save_gradients,
        detach_saved_tensors,
        mark_input_output_distances,
    )
    model_history._run_and_log_inputs_through_model(
        model, input_args, input_kwargs, layers_to_save, random_seed
    )
    return model_history


def log_forward_pass(
    model: nn.Module,
    input_args: Union[torch.Tensor, List[Any], Tuple[Any]],
    input_kwargs: Dict[Any, Any] = None,
    layers_to_save: Optional[Union[str, List]] = "all",
    keep_unsaved_layers: bool = True,
    output_device: str = "same",
    activation_postfunc: Optional[Callable] = None,
    mark_input_output_distances: bool = False,
    detach_saved_tensors: bool = False,
    save_function_args: bool = False,
    save_gradients: bool = False,
    vis_opt: str = "none",
    vis_nesting_depth: int = 1000,
    vis_outpath: str = "graph.gv",
    vis_save_only: bool = False,
    vis_fileformat: str = "pdf",
    vis_buffer_layers: bool = False,
    vis_direction: str = "bottomup",
    vis_graph_overrides: Dict = None,
    vis_node_overrides: Dict = None,
    vis_nested_node_overrides: Dict = None,
    vis_edge_overrides: Dict = None,
    vis_gradient_edge_overrides: Dict = None,
    vis_module_overrides: Dict = None,
    random_seed: Optional[int] = None,
) -> ModelHistory:
    """Runs a forward pass through a model given input x, and returns a ModelHistory object containing a log
    (layer activations and accompanying layer metadata) of the forward pass for all layers specified in which_layers,
    and optionally visualizes the model graph if vis_opt is set to 'rolled' or 'unrolled'.

    In which_layers, can specify 'all', for all layers (default), or a list containing any combination of:
    1) desired layer names (e.g., 'conv2d_1_1'; if a layer has multiple passes, this includes all passes),
    2) a layer pass (e.g., conv2d_1_1:2 for just the second pass), 3) a module name to fetch the output of a particular
    module, 4) the ordinal index of a layer in the model (e.g. 3 for the third layer, -2 for the second to last, etc.),
    or 5) a desired substring with which to filter desired layers (e.g., 'conv2d' for all conv2d layers).

    Args:
        model: PyTorch model
        input_args: input arguments for model forward pass; as a list if multiple, else as a single tensor.
        input_kwargs: keyword arguments for model forward pass
        layers_to_save: list of layers to include (described above), or 'all' to include all layers.
        keep_unsaved_layers: whether to keep layers without saved activations in the log (i.e., with just metadata)
        activation_postfunc: Function to apply to tensors before saving them (e.g., channelwise averaging).
        mark_input_output_distances: whether to mark the distance of each layer from the input or output;
            False by default since this is computationally expensive.
        output_device: device where saved tensors are to be stored. Either 'same' to keep on the same device,
            or 'cpu' or 'cuda' to move them to cpu or cuda when saved.
        detach_saved_tensors: whether to detach the saved tensors, so they remain attached to the computational graph
        save_function_args: whether to save the arguments to each function involved in computing each tensor
        save_gradients: whether to save gradients from any subsequent backward pass
        vis_opt: whether, and how, to visualize the network; 'none' for
            no visualization, 'rolled' to show the graph in rolled-up format (i.e.,
            one node per layer if a recurrent network), or 'unrolled' to show the graph
            in unrolled format (i.e., one node per pass through a layer if a recurrent)
        vis_nesting_depth: How many levels of nested modules to show; 1 for only top-level modules, 2 for two
            levels, etc.
        vis_outpath: file path to save the graph visualization
        vis_save_only: whether to only save the graph visual without immediately showing it
        vis_fileformat: the format of the visualization (e.g,. 'pdf', 'jpg', etc.)
        vis_buffer_layers: whether to visualize the buffer layers
        vis_direction: either 'bottomup', 'topdown', or 'leftright'
        random_seed: which random seed to use in case model involves randomness

    Returns:
        ModelHistory object with layer activations and metadata
    """
    warn_parallel()

    if vis_opt not in ["none", "rolled", "unrolled"]:
        raise ValueError("Visualization option must be either 'none', 'rolled', or 'unrolled'.")

    if output_device not in ["same", "cpu", "cuda"]:
        raise ValueError("output_device must be either 'same', 'cpu', or 'cuda'.")

    if type(layers_to_save) is str:
        layers_to_save = layers_to_save.lower()

    if layers_to_save in ["all", "none", None, []]:
        model_history = run_model_and_save_specified_activations(
            model=model,
            input_args=input_args,
            input_kwargs=input_kwargs,
            layers_to_save=layers_to_save,
            keep_unsaved_layers=keep_unsaved_layers,
            output_device=output_device,
            activation_postfunc=activation_postfunc,
            mark_input_output_distances=mark_input_output_distances,
            detach_saved_tensors=detach_saved_tensors,
            save_function_args=save_function_args,
            save_gradients=save_gradients,
            random_seed=random_seed,
        )
    else:
        model_history = run_model_and_save_specified_activations(
            model=model,
            input_args=input_args,
            input_kwargs=input_kwargs,
            layers_to_save=None,
            keep_unsaved_layers=True,
            output_device=output_device,
            activation_postfunc=activation_postfunc,
            mark_input_output_distances=mark_input_output_distances,
            detach_saved_tensors=detach_saved_tensors,
            save_function_args=save_function_args,
            save_gradients=save_gradients,
            random_seed=random_seed,
        )
        model_history.keep_unsaved_layers = keep_unsaved_layers
        model_history.save_new_activations(
            model=model,
            input_args=input_args,
            input_kwargs=input_kwargs,
            layers_to_save=layers_to_save,
            random_seed=random_seed,
        )

    # Visualize if desired.
    if vis_opt != "none":
        model_history.render_graph(
            vis_opt,
            vis_nesting_depth,
            vis_outpath,
            vis_graph_overrides,
            vis_node_overrides,
            vis_nested_node_overrides,
            vis_edge_overrides,
            vis_gradient_edge_overrides,
            vis_module_overrides,
            vis_save_only,
            vis_fileformat,
            vis_buffer_layers,
            vis_direction,
        )

    return model_history


def get_model_metadata(
    model: nn.Module,
    input_args: Union[torch.Tensor, List[Any], Tuple[Any]],
    input_kwargs: Dict[Any, Any] = None,
) -> ModelHistory:
    """Logs all metadata for a given model and inputs without saving any activations. NOTE: this function
    will be removed in a future version of TorchLens, since calling it is identical to calling
    log_forward_pass without saving any layers.

    Args:
        model: model to inspect
        input_args: list of input positional arguments, or a single tensor
        input_kwargs: dict of keyword arguments
    Returns:
        ModelHistory object with metadata about the model.
    """
    model_history = log_forward_pass(
        model,
        input_args,
        input_kwargs,
        layers_to_save=None,
        mark_input_output_distances=True,
    )
    return model_history


def show_model_graph(
    model: nn.Module,
    input_args: Union[torch.Tensor, List, Tuple],
    input_kwargs: Dict[Any, Any] = None,
    vis_opt: str = "unrolled",
    vis_nesting_depth: int = 1000,
    vis_outpath: str = "graph.gv",
    vis_graph_overrides: Dict = None,
    vis_node_overrides: Dict = None,
    vis_nested_node_overrides: Dict = None,
    vis_edge_overrides: Dict = None,
    vis_gradient_edge_overrides: Dict = None,
    vis_module_overrides: Dict = None,
    save_only: bool = False,
    vis_fileformat: str = "pdf",
    vis_buffer_layers: bool = False,
    vis_direction: str = "bottomup",
    random_seed: Optional[int] = None,
) -> None:
    """Visualize the model graph without saving any activations.

    Args:
        model: PyTorch model
        input_args: Arguments for model forward pass
        input_kwargs: Keyword arguments for model forward pass
        vis_opt: whether, and how, to visualize the network; 'none' for
            no visualization, 'rolled' to show the graph in rolled-up format (i.e.,
            one node per layer if a recurrent network), or 'unrolled' to show the graph
            in unrolled format (i.e., one node per pass through a layer if a recurrent)
        vis_nesting_depth: How many levels of nested modules to show; 1 for only top-level modules, 2 for two
            levels, etc.
        vis_outpath: file path to save the graph visualization
        save_only: whether to only save the graph visual without immediately showing it
        vis_fileformat: the format of the visualization (e.g,. 'pdf', 'jpg', etc.)
        vis_buffer_layers: whether to visualize the buffer layers
        vis_direction: either 'bottomup', 'topdown', or 'leftright'
        random_seed: which random seed to use in case model involves randomness

    Returns:
        Nothing.
    """
    if not input_kwargs:
        input_kwargs = {}

    if vis_opt not in ["none", "rolled", "unrolled"]:
        raise ValueError("Visualization option must be either 'none', 'rolled', or 'unrolled'.")

    model_history = run_model_and_save_specified_activations(
        model=model,
        input_args=input_args,
        input_kwargs=input_kwargs,
        layers_to_save=None,
        activation_postfunc=None,
        mark_input_output_distances=False,
        detach_saved_tensors=False,
        save_gradients=False,
        random_seed=random_seed,
    )
    model_history.render_graph(
        vis_opt,
        vis_nesting_depth,
        vis_outpath,
        vis_graph_overrides,
        vis_node_overrides,
        vis_nested_node_overrides,
        vis_edge_overrides,
        vis_gradient_edge_overrides,
        vis_module_overrides,
        save_only,
        vis_fileformat,
        vis_buffer_layers,
        vis_direction,
    )


def validate_saved_activations(
    model: nn.Module,
    input_args: Union[torch.Tensor, List[Any], Tuple[Any]],
    input_kwargs: Dict[Any, Any] = None,
    random_seed: Union[int, None] = None,
    verbose: bool = False,
) -> bool:
    """Validate that the saved model activations correctly reproduce the ground truth output. This function works by
    running a forward pass through the model, saving all activations, re-running the forward pass starting from
    the saved activations in each layer, and checking that the resulting output matches the original output.
    Additionally, it substitutes in random activations and checks whether the output changes accordingly, for
    at least min_proportion_consequential_layers of the layers (in case some layers do not change the output for some
    reason). Returns True if a model passes these tests for the given input, and False otherwise.

    Args:
        model: PyTorch model.
        input_args: Input for which to validate the saved activations.
        input_kwargs: Keyword arguments for model forward pass
        random_seed: random seed in case model is stochastic
        verbose: whether to show verbose error messages
    Returns:
        True if the saved activations correctly reproduce the ground truth output, false otherwise.
    """
    warn_parallel()
    if random_seed is None:  # set random seed
        random_seed = random.randint(1, 4294967294)
    set_random_seed(random_seed)
    if type(input_args) is tuple:
        input_args = list(input_args)
    elif (type(input_args) not in [list, tuple]) and (input_args is not None):
        input_args = [input_args]
    if not input_args:
        input_args = []
    if not input_kwargs:
        input_kwargs = {}
    input_args_copy = [copy.deepcopy(arg) for arg in input_args]
    input_kwargs_copy = {key: copy.deepcopy(val) for key, val in input_kwargs.items()}
    state_dict = model.state_dict()
    ground_truth_output_all = get_vars_of_type_from_obj(
        model(*input_args_copy, **input_kwargs_copy),
        torch.Tensor,
        search_depth=5,
        return_addresses=True,
        allow_repeats=True,
    )
    # Deduplicate by address to match trace_model.py output extraction
    addresses_used = []
    ground_truth_output_tensors = []
    for entry in ground_truth_output_all:
        if entry[1] in addresses_used:
            continue
        ground_truth_output_tensors.append(entry[0])
        addresses_used.append(entry[1])
    model.load_state_dict(state_dict)
    model_history = run_model_and_save_specified_activations(
        model=model,
        input_args=input_args,
        input_kwargs=input_kwargs,
        layers_to_save="all",
        keep_unsaved_layers=True,
        activation_postfunc=None,
        mark_input_output_distances=False,
        detach_saved_tensors=False,
        save_gradients=False,
        save_function_args=True,
        random_seed=random_seed,
    )
    activations_are_valid = model_history.validate_saved_activations(
        ground_truth_output_tensors, verbose
    )

    model_history.cleanup()
    del model_history
    return activations_are_valid


def validate_batch_of_models_and_inputs(
    models_and_inputs_dict: Dict[str, Dict[str, Union[str, Callable, Dict]]],
    out_path: str,
    redo_model_if_already_run: bool = True,
) -> pd.DataFrame:
    """Given multiple models and several inputs for each, validates the saved activations for all of them
    and returns a Pandas dataframe summarizing the validation results.

    Args:
        models_and_inputs_dict: Dict mapping each model name to a dict of model info:
                model_category: category of model (e.g., torchvision; just for directory bookkeeping)
                model_loading_func: function to load the model
                model_sample_inputs: dict of example inputs {name: input}
        out_path: Path to save the validation results to
        redo_model_if_already_run: If True, will re-run the validation for a model even if already run

    Returns:
        Pandas dataframe with validation information for each model and input
    """
    if os.path.exists(out_path):
        current_csv = pd.read_csv(out_path)
    else:
        current_csv = pd.DataFrame.from_dict(
            {
                "model_category": [],
                "model_name": [],
                "input_name": [],
                "validation_success": [],
            }
        )
    models_already_run = current_csv["model_name"].unique()
    for model_name, model_info in tqdm(models_and_inputs_dict.items(), desc="Validating models"):
        print(f"Validating model {model_name}")
        if model_name in models_already_run and not redo_model_if_already_run:
            continue
        model_category = model_info["model_category"]
        model_loading_func = model_info["model_loading_func"]
        model = model_loading_func()
        model_sample_inputs = model_info["model_sample_inputs"]
        for input_name, x in model_sample_inputs.items():
            validation_success = validate_saved_activations(model, x)
            current_csv = pd.concat(
                [
                    current_csv,
                    pd.DataFrame(
                        [
                            {
                                "model_category": model_category,
                                "model_name": model_name,
                                "input_name": input_name,
                                "validation_success": validation_success,
                            }
                        ]
                    ),
                ],
                ignore_index=True,
            )
        current_csv.to_csv(out_path, index=False)
        del model
    return current_csv



"""Tests for real-world models.

All optional-dependency imports are local (inside test functions) using
pytest.importorskip() for non-torchvision packages. Tests with missing
packages show as SKIPPED, never ERROR.

Only torch, torchvision, pytest, and torchlens are imported at the top level.
"""

from os.path import join as opj

import numpy as np
import pytest
import torch
import torchvision

from conftest import VIS_OUTPUT_DIR

import example_models
from torchlens import show_model_graph, validate_saved_activations


# =============================================================================
# TorchVision Classification Models
# =============================================================================


def test_alexnet(default_input1):
    model = torchvision.models.AlexNet()
    assert validate_saved_activations(model, default_input1)
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "alexnet"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=1,
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "alexnet_depth1"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=2,
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "alexnet_depth2"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=3,
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "alexnet_depth3"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=4,
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "alexnet_depth4"),
    )


def test_vgg16(default_input1):
    model = torchvision.models.vgg16()
    assert validate_saved_activations(model, default_input1)
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "vgg16"),
    )


def test_vit(default_input1):
    model = torchvision.models.vit_l_16()
    assert validate_saved_activations(model, default_input1)
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchvision-main", "vit_l_16"),
    )


# =============================================================================
# CORNet Models (requires cornet package)
# =============================================================================


def test_cornet_z(default_input1):
    cornet = pytest.importorskip("cornet")
    model = cornet.cornet_z()
    assert validate_saved_activations(model, default_input1)
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_unrolled"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="rolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_rolled"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=1,
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_unrolled_depth1"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="rolled",
        vis_nesting_depth=1,
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_rolled_depth1"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=2,
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_unrolled_depth2"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="rolled",
        vis_nesting_depth=2,
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_rolled_depth2"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_nesting_depth=3,
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_unrolled_depth3"),
    )
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="rolled",
        vis_nesting_depth=3,
        vis_outpath=opj(VIS_OUTPUT_DIR, "cornet", "cornet_z_rolled_depth3"),
    )


# =============================================================================
# TIMM Models (requires timm)
# =============================================================================


def test_timm_beit_base_patch16_224(default_input1):
    timm = pytest.importorskip("timm")
    model = timm.models.beit_base_patch16_224()
    show_model_graph(
        model,
        default_input1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "timm", "timm_beit_base_patch16_224"),
    )
    assert validate_saved_activations(model, default_input1)


def test_timm_gluon_resnext101_32x4d():
    timm = pytest.importorskip("timm")
    model = timm.create_model("gluon_resnext101_32x4d", pretrained=False)
    model_input = torch.randn(1, 3, 224, 224)
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "timm", "gluon_resnext101_32x4d"),
    )
    assert validate_saved_activations(model, model_input)


def test_timm_ecaresnet101d():
    timm = pytest.importorskip("timm")
    model = timm.create_model("ecaresnet101d", pretrained=False)
    model_input = torch.randn(1, 3, 224, 224)
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "timm", "ecaresnet101d"),
    )
    assert validate_saved_activations(model, model_input)


# =============================================================================
# Torchaudio Models (requires torchaudio)
# =============================================================================


def test_audio_conv_tasnet_base():
    torchaudio = pytest.importorskip("torchaudio")
    model = torchaudio.models.conv_tasnet_base()
    model_input = torch.randn(6, 1, 3)
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchaudio", "audio_conv_tasnet_base"),
    )
    assert validate_saved_activations(model, model_input)


def test_audio_wav2letter():
    torchaudio = pytest.importorskip("torchaudio")
    model = torchaudio.models.Wav2Letter()
    model_input = torch.randn(1, 1, 2000)
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "torchaudio", "audio_wav2letter"),
    )
    assert validate_saved_activations(model, model_input)


# =============================================================================
# Language Models
# =============================================================================


def test_lstm():
    model = example_models.LSTMModel()
    model_input = torch.rand(5, 5, 5)
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "language-models", "language_lstm_unrolled"),
    )
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="rolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "language-models", "language_lstm_rolled"),
    )
    assert validate_saved_activations(model, model_input)


def test_rnn():
    model = example_models.RNNModel()
    model_input = torch.rand(5, 5, 5)
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "language-models", "language_rnn_unrolled"),
    )
    show_model_graph(
        model,
        model_input,
        save_only=True,
        vis_opt="rolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "language-models", "language_rnn_rolled"),
    )
    assert validate_saved_activations(model, model_input)


def test_gpt2():
    transformers = pytest.importorskip("transformers")
    tokenizer = transformers.GPT2Tokenizer.from_pretrained("gpt2")
    model = transformers.GPT2Model.from_pretrained("gpt2")
    model_inputs = ["to be or not to be", "that is the question"]
    model_inputs = tokenizer(*model_inputs, return_tensors="pt")
    show_model_graph(
        model,
        [],
        model_inputs,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "language-models", "gpt2"),
    )
    assert validate_saved_activations(model, [], model_inputs)


# =============================================================================
# Multimodal / Special Models
# =============================================================================


def test_stable_diffusion():
    model = example_models.ContextUnet(3, 16, 10)
    model_inputs = (
        torch.rand(1, 3, 28, 28),
        torch.tensor([1]),
        torch.tensor([1.0]),
        torch.tensor([3.0]),
    )
    show_model_graph(
        model,
        model_inputs,
        random_seed=1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "multimodal-models", "stable_diffusion"),
    )
    assert validate_saved_activations(model, model_inputs, random_seed=1)


def test_styletts():
    try:
        from StyleTTS.models import TextEncoder
    except ModuleNotFoundError:
        pytest.skip("StyleTTS not available")

    model = TextEncoder(3, 3, 3, 100)
    model.eval()
    tokens = torch.tensor([[3, 0, 1, 2, 0, 2, 2, 3, 1, 4]])
    input_lengths = torch.ones(1, dtype=torch.long) * 10
    m = torch.ones(1, 10, dtype=torch.bool)
    model_inputs = (tokens, input_lengths, m)
    show_model_graph(
        model,
        model_inputs,
        random_seed=1,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "text-to-speech", "styletts_text_encoder"),
    )
    assert validate_saved_activations(model, model_inputs, random_seed=1)


def test_qml():
    qml = pytest.importorskip("pennylane")
    n_qubits = 2
    dev = qml.device("default.qubit", wires=n_qubits)

    @qml.qnode(dev, diff_method="backprop")
    def qnode(inputs, weights):
        qml.RX(inputs[0][0], wires=0)
        qml.RY(weights, wires=0)
        return [qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliZ(1))]

    weight_shapes = {"weights": 1}
    qlayer = qml.qnn.TorchLayer(qnode, weight_shapes)

    clayer_1 = torch.nn.Linear(2, 2)
    clayer_2 = torch.nn.Linear(2, 2)
    softmax = torch.nn.Softmax(dim=1)
    layers = [clayer_1, qlayer, clayer_2, softmax]
    model = torch.nn.Sequential(*layers)
    model_inputs = torch.rand(1, 2, requires_grad=False)

    show_model_graph(
        model,
        model_inputs,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "quantum", "qml"),
    )
    assert validate_saved_activations(model, model_inputs)


def test_lightning():
    L = pytest.importorskip("lightning")
    import torch.nn as nn

    class OneHotAutoEncoder(L.LightningModule):
        def __init__(self):
            super().__init__()
            self.model = nn.Sequential(
                nn.Linear(16, 4),
                nn.ReLU(),
                nn.Linear(4, 16),
            )

        def forward(self, x):
            x_hat = self.model(x)
            return nn.functional.mse_loss(x_hat, x)

    model = OneHotAutoEncoder()
    model_inputs = [torch.randn(2, 16)]

    show_model_graph(
        model,
        model_inputs,
        save_only=True,
        vis_opt="unrolled",
        vis_outpath=opj(VIS_OUTPUT_DIR, "lightning", "one-hot-autoencoder"),
    )
    assert validate_saved_activations(model, model_inputs)



# This file is for defining the ModelHistory class that stores the representation of the forward pass.
import copy
from collections import OrderedDict, defaultdict
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from .cleanup import _remove_log_entry, cleanup
from .decorate_torch import decorate_pytorch
from .helper_funcs import (
    human_readable_size,
)
from .interface import (
    _getitem_after_pass,
    _getitem_during_pass,
    _str_after_pass,
    _str_during_pass,
    to_pandas,
    print_all_fields,
)
from .logging_funcs import save_new_activations
from .model_funcs import cleanup_model, prepare_model
from .postprocess import postprocess
from .tensor_log import RolledTensorLogEntry, TensorLogEntry
from .trace_model import run_and_log_inputs_through_model
from .validation import validate_saved_activations
from .vis import render_graph


class ModelHistory:
    def __init__(
        self,
        model_name: str,
        output_device: str = "same",
        activation_postfunc: Optional[Callable] = None,
        keep_unsaved_layers: bool = True,
        save_function_args: bool = False,
        save_gradients: bool = False,
        detach_saved_tensors: bool = False,
        mark_input_output_distances: bool = True,
    ):
        """Object that stores the history of a model's forward pass.
        Both logs the history in real time, and stores a nice
        representation of the full history for the user afterward.
        """
        # Setup:
        activation_postfunc = copy.deepcopy(activation_postfunc)

        # General info
        self.model_name = model_name
        self._pass_finished = False
        self._track_tensors = False
        self.logging_mode = "exhaustive"
        self._pause_logging = False
        self._all_layers_logged = False
        self._all_layers_saved = False
        self.keep_unsaved_layers = keep_unsaved_layers
        self.activation_postfunc = activation_postfunc
        self.current_function_call_barcode = None
        self.random_seed_used = None
        self.output_device = output_device
        self.detach_saved_tensors = detach_saved_tensors
        self.save_function_args = save_function_args
        self.save_gradients = save_gradients
        self.has_saved_gradients = False
        self.mark_input_output_distances = mark_input_output_distances

        # Model structure info
        self.model_is_recurrent = False
        self.model_max_recurrent_loops = 1
        self.model_has_conditional_branching = False
        self.model_is_branching = False

        # Tensor Tracking:
        self.layer_list: List[TensorLogEntry] = []
        self.layer_list_rolled: List[RolledTensorLogEntry] = []
        self.layer_dict_main_keys: Dict[str, TensorLogEntry] = OrderedDict()
        self.layer_dict_all_keys: Dict[str, TensorLogEntry] = OrderedDict()
        self.layer_dict_rolled: Dict[str, RolledTensorLogEntry] = OrderedDict()
        self.layer_labels: List[str] = []
        self.layer_labels_w_pass: List[str] = []
        self.layer_labels_no_pass: List[str] = []
        self.layer_num_passes: Dict[str, int] = OrderedDict()
        self._raw_tensor_dict: Dict[str, TensorLogEntry] = OrderedDict()
        self._raw_tensor_labels_list: List[str] = []
        self._tensor_nums_to_save: List[int] = []
        self._tensor_counter: int = 0
        self.num_operations: int = 0
        self._raw_layer_type_counter: Dict[str, int] = defaultdict(lambda: 0)
        self._unsaved_layers_lookup_keys: Set[str] = set()

        # Mapping from raw to final layer labels:
        self._raw_to_final_layer_labels: Dict[str, str] = {}
        self._final_to_raw_layer_labels: Dict[str, str] = {}
        self._lookup_keys_to_tensor_num_dict: Dict[str, int] = {}
        self._tensor_num_to_lookup_keys_dict: Dict[int, List[str]] = defaultdict(list)

        # Special Layers:
        self.input_layers: List[str] = []
        self.output_layers: List[str] = []
        self.buffer_layers: List[str] = []
        self.buffer_num_passes: Dict = {}
        self.internally_initialized_layers: List[str] = []
        self._layers_where_internal_branches_merge_with_input: List[str] = []
        self.internally_terminated_layers: List[str] = []
        self.internally_terminated_bool_layers: List[str] = []
        self.conditional_branch_edges: List[Tuple[str, str]] = []
        self.layers_with_saved_activations: List[str] = []
        self.orphan_layers: List[str] = []
        self.unlogged_layers: List[str] = []
        self.layers_with_saved_gradients: List[str] = []
        self.layers_computed_with_params: Dict[str, List] = defaultdict(list)
        self.equivalent_operations: Dict[str, set] = defaultdict(set)

        # Tensor info:
        self.num_tensors_total: int = 0
        self.tensor_fsize_total: int = 0
        self.tensor_fsize_total_nice: str = human_readable_size(0)
        self.num_tensors_saved: int = 0
        self.tensor_fsize_saved: int = 0
        self.tensor_fsize_saved_nice: str = human_readable_size(0)

        # Param info:
        self.total_param_tensors: int = 0
        self.total_param_layers: int = 0
        self.total_params: int = 0
        self.total_params_fsize: int = 0
        self.total_params_fsize_nice: str = human_readable_size(0)

        # Module info:
        self.module_addresses: List[str] = []
        self.module_types: Dict[str, Any] = {}
        self.module_passes: List = []
        self.module_num_passes: Dict = defaultdict(lambda: 1)
        self.top_level_modules: List = []
        self.top_level_module_passes: List = []
        self.module_children: Dict = defaultdict(list)
        self.module_pass_children: Dict = defaultdict(list)
        self.module_nparams: Dict = defaultdict(lambda: 0)
        self.module_num_tensors: Dict = defaultdict(lambda: 0)
        self.module_pass_num_tensors: Dict = defaultdict(lambda: 0)
        self.module_layers: Dict = defaultdict(list)
        self.module_pass_layers: Dict = defaultdict(list)
        self.module_layer_argnames = defaultdict(list)

        # Time elapsed:
        self.pass_start_time: float = 0
        self.pass_end_time: float = 0
        self.elapsed_time_setup: float = 0
        self.elapsed_time_forward_pass: float = 0
        self.elapsed_time_cleanup: float = 0
        self.elapsed_time_total: float = 0
        self.elapsed_time_function_calls: float = 0
        self.elapsed_time_torchlens_logging: float = 0

        # Reference info
        self.func_argnames: Dict[str, tuple] = defaultdict(lambda: tuple([]))

    # ********************************************
    # ************ Built-in Methods **************
    # ********************************************

    def __len__(self):
        if self._pass_finished:
            return len(self.layer_list)
        else:
            return len(self._raw_tensor_dict)

    def __getitem__(self, ix) -> TensorLogEntry:
        """Returns an object logging a model layer given an index. If the pass is finished,
        it'll do this intelligently; if not, it simply queries based on the layer's raw barcode.

        Args:
            ix: desired index

        Returns:
            Tensor log entry object with info about specified layer.
        """
        if self._pass_finished:
            return _getitem_after_pass(self, ix)
        else:
            return _getitem_during_pass(self, ix)

    def __str__(self) -> str:
        if self._pass_finished:
            return _str_after_pass(self)
        else:
            return _str_during_pass(self)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        """Loops through all tensors in the log."""
        if self._pass_finished:
            return iter(self.layer_list)
        else:
            return iter(list(self._raw_tensor_dict.values()))

    # ********************************************
    # ************* FLOPs Properties *************
    # ********************************************

    @property
    def total_flops_forward(self) -> int:
        """Total forward FLOPs across all layers (skipping None/unknown)."""
        return sum(
            entry.flops_forward for entry in self.layer_list if entry.flops_forward is not None
        )

    @property
    def total_flops_backward(self) -> int:
        """Total backward FLOPs across all layers (skipping None/unknown)."""
        return sum(
            entry.flops_backward for entry in self.layer_list if entry.flops_backward is not None
        )

    @property
    def total_flops(self) -> int:
        """Total FLOPs (forward + backward)."""
        return self.total_flops_forward + self.total_flops_backward

    def flops_by_type(self) -> Dict[str, Dict[str, int]]:
        """Group FLOPs by layer type.

        Returns:
            Dict mapping layer_type to {"forward": int, "backward": int, "count": int}.
        """
        result: Dict[str, Dict[str, int]] = {}
        for entry in self.layer_list:
            lt = entry.layer_type
            if lt not in result:
                result[lt] = {"forward": 0, "backward": 0, "count": 0}
            result[lt]["count"] += 1
            if entry.flops_forward is not None:
                result[lt]["forward"] += entry.flops_forward
            if entry.flops_backward is not None:
                result[lt]["backward"] += entry.flops_backward
        return result

    # ********************************************
    # ******** Assign Imported Methods ***********
    # ********************************************

    render_graph = render_graph
    print_all_fields = print_all_fields
    to_pandas = to_pandas
    save_new_activations = save_new_activations
    validate_saved_activations = validate_saved_activations
    cleanup = cleanup
    _postprocess = postprocess
    _decorate_pytorch = decorate_pytorch
    _prepare_model = prepare_model
    _cleanup_model = cleanup_model
    _run_and_log_inputs_through_model = run_and_log_inputs_through_model
    _remove_log_entry = _remove_log_entry



import warnings
from functools import wraps
from typing import Callable, Dict, List, TYPE_CHECKING

import torch
from torch import nn

from .helper_funcs import (
    get_vars_of_type_from_obj,
    iter_accessible_attributes,
    remove_attributes_starting_with_str,
)
from .logging_funcs import log_source_tensor

if TYPE_CHECKING:
    from .model_history import ModelHistory


def prepare_model(
    model_log: "ModelHistory",
    model: nn.Module,
    module_orig_forward_funcs: Dict,
    decorated_func_mapper: Dict[Callable, Callable],
):
    """Adds annotations and hooks to the model, and decorates any functions in the model.

    Args:
        model: Model to prepare.
        module_orig_forward_funcs: Dict with the original forward funcs for each submodule
        decorated_func_mapper: Dictionary mapping decorated functions to original functions, so they can be restored

    Returns:
        Model with hooks and attributes added.
    """
    model_log.model_name = str(type(model).__name__)
    model.tl_module_address = ""
    model.tl_source_model_history = model_log

    module_stack = [(model, "")]  # list of tuples (name, module)

    while len(module_stack) > 0:
        module, parent_address = module_stack.pop()
        module_children = list(module.named_children())

        # Decorate any torch functions in the model:
        for func_name, func in module.__dict__.items():
            if (
                (func_name[0:2] == "__")
                or (not callable(func))
                or (func not in decorated_func_mapper)
            ):
                continue
            module.__dict__[func_name] = decorated_func_mapper[func]

        # Annotate the children with the full address.
        for c, (child_name, child_module) in enumerate(module_children):
            child_address = f"{parent_address}.{child_name}" if parent_address != "" else child_name
            child_module.tl_module_address = child_address
            module_children[c] = (child_module, child_address)
        module_stack = module_children + module_stack

        if module == model:  # don't tag the model itself.
            continue

        module.tl_source_model_history = model_log
        module.tl_module_type = str(type(module).__name__)
        model_log.module_types[module.tl_module_address] = module.tl_module_type
        module.tl_module_pass_num = 0
        module.tl_module_pass_labels = []
        module.tl_tensors_entered_labels = []
        module.tl_tensors_exited_labels = []

        # Add decorators.

        if hasattr(module, "forward") and not hasattr(
            module.forward, "tl_forward_call_is_decorated"
        ):
            module_orig_forward_funcs[module] = module.forward
            module.forward = module_forward_decorator(model_log, module.forward, module)
            module.forward.tl_forward_call_is_decorated = True

    # Mark all parameters with requires_grad = True, and mark what they were before, so they can be restored on cleanup.
    for param in model.parameters():
        param.tl_requires_grad = param.requires_grad
        param.requires_grad = True

    # And prepare any buffer tensors.
    prepare_buffer_tensors(model_log, model)


def prepare_buffer_tensors(model_log, model: nn.Module):
    """Goes through a model and all its submodules, and prepares any "buffer" tensors: tensors
    attached to the module that aren't model parameters.

    Args:
        model: PyTorch model

    Returns:
        PyTorch model with all buffer tensors prepared and ready to track.
    """
    submodules = get_all_submodules(model)
    for submodule in submodules:
        attr_list = list(submodule.named_buffers()) + list(iter_accessible_attributes(submodule))
        for attribute_name, attribute in attr_list:
            if (
                issubclass(type(attribute), torch.Tensor)
                and not issubclass(type(attribute), torch.nn.Parameter)
                and not hasattr(attribute, "tl_buffer_address")
            ):
                if submodule.tl_module_address == "":
                    buffer_address = attribute_name
                else:
                    buffer_address = submodule.tl_module_address + "." + attribute_name
                setattr(attribute, "tl_buffer_address", buffer_address)


def module_forward_decorator(model_log, orig_forward: Callable, module: nn.Module) -> Callable:
    @wraps(orig_forward)
    def decorated_forward(*args, **kwargs):
        if model_log.logging_mode == "fast":  # do bare minimum for logging.
            out = orig_forward(*args, **kwargs)
            input_tensors_fast = get_vars_of_type_from_obj(
                [args, kwargs], torch.Tensor, [torch.nn.Parameter], search_depth=5
            )
            input_tensor_labels = {
                t.tl_tensor_label_raw
                for t in input_tensors_fast
                if hasattr(t, "tl_tensor_label_raw")
            }
            output_tensors = get_vars_of_type_from_obj(out, torch.Tensor, search_depth=4)
            for t in output_tensors:
                if (module.tl_module_type.lower() == "identity") or (
                    hasattr(t, "tl_tensor_label_raw")
                    and t.tl_tensor_label_raw in input_tensor_labels
                ):
                    t = getattr(torch, "identity")(t)
            return out

        # "Pre-hook" operations:
        module_address = module.tl_module_address
        module.tl_module_pass_num += 1
        module_pass_label = (module_address, module.tl_module_pass_num)
        module.tl_module_pass_labels.append(module_pass_label)
        input_tensors = get_vars_of_type_from_obj(
            [args, kwargs], torch.Tensor, [torch.nn.Parameter], search_depth=5
        )
        input_tensor_labels = set()
        for t in input_tensors:
            if (not hasattr(t, "tl_tensor_label_raw")) and hasattr(t, "tl_buffer_address"):
                log_source_tensor(model_log, t, "buffer", getattr(t, "tl_buffer_address"))
            tensor_entry = model_log._raw_tensor_dict[t.tl_tensor_label_raw]
            input_tensor_labels.add(t.tl_tensor_label_raw)
            module.tl_tensors_entered_labels.append(t.tl_tensor_label_raw)
            tensor_entry.modules_entered.append(module_address)
            tensor_entry.module_passes_entered.append(module_pass_label)
            tensor_entry.is_submodule_input = True
            for arg_key, arg_val in list(enumerate(args)) + list(kwargs.items()):
                if arg_val is t:
                    tensor_entry.modules_entered_argnames[
                        f"{module_pass_label[0]}:{module_pass_label[1]}"
                    ].append(arg_key)
                    model_log.module_layer_argnames[
                        (f"{module_pass_label[0]}:{module_pass_label[1]}")
                    ].append((t.tl_tensor_label_raw, arg_key))
            tensor_entry.module_entry_exit_thread_output.append(
                ("+", module_pass_label[0], module_pass_label[1])
            )

        # Check the buffers.
        for buffer_name, buffer_tensor in module.named_buffers():
            if hasattr(buffer_tensor, "tl_buffer_address"):
                continue
            if module.tl_module_address == "":
                buffer_address = buffer_name
            else:
                buffer_address = f"{module.tl_module_address}.{buffer_name}"
            buffer_tensor.tl_buffer_address = buffer_address
            if hasattr(buffer_tensor, "tl_tensor_label_raw"):
                buffer_tensor.tl_buffer_parent = buffer_tensor.tl_tensor_label_raw
                delattr(buffer_tensor, "tl_tensor_label_raw")

        # The function call
        out = orig_forward(*args, **kwargs)

        # "Post-hook" operations:
        module_address = module.tl_module_address
        module_pass_num = module.tl_module_pass_num
        module_entry_label = module.tl_module_pass_labels.pop()
        output_tensors = get_vars_of_type_from_obj(out, torch.Tensor, search_depth=4)
        for t in output_tensors:
            # if identity module or tensor unchanged, run the identity function for bookkeeping
            if (module.tl_module_type.lower() == "identity") or (
                t.tl_tensor_label_raw in input_tensor_labels
            ):
                t = getattr(torch, "identity")(t)
            tensor_entry = model_log._raw_tensor_dict[t.tl_tensor_label_raw]
            tensor_entry.is_submodule_output = True
            tensor_entry.is_bottom_level_submodule_output = (
                log_whether_exited_submodule_is_bottom_level(model_log, t, module)
            )
            tensor_entry.modules_exited.append(module_address)
            tensor_entry.module_passes_exited.append((module_address, module_pass_num))
            tensor_entry.module_entry_exit_thread_output.append(
                ("-", module_entry_label[0], module_entry_label[1])
            )
            module.tl_tensors_exited_labels.append(t.tl_tensor_label_raw)

        for t in (
            input_tensors
        ):  # Now that module is finished, roll back the threads of all input tensors.
            tensor_entry = model_log._raw_tensor_dict[t.tl_tensor_label_raw]
            input_module_thread = tensor_entry.module_entry_exit_thread_output[:]
            if (
                "+",
                module_entry_label[0],
                module_entry_label[1],
            ) in input_module_thread:
                module_entry_ix = input_module_thread.index(
                    ("+", module_entry_label[0], module_entry_label[1])
                )
                tensor_entry.module_entry_exit_thread_output = (
                    tensor_entry.module_entry_exit_thread_output[:module_entry_ix]
                )

        return out

    return decorated_forward


def log_whether_exited_submodule_is_bottom_level(model_log, t: torch.Tensor, submodule: nn.Module):
    """Checks whether the submodule that a tensor is leaving is a "bottom-level" submodule;
    that is, that only one tensor operation happened inside the submodule.

    Args:
        t: the tensor leaving the module
        submodule: the module that the tensor is leaving

    Returns:
        Whether the tensor operation is bottom level.
    """
    tensor_entry = model_log._raw_tensor_dict[getattr(t, "tl_tensor_label_raw")]
    submodule_address = submodule.tl_module_address

    if tensor_entry.is_bottom_level_submodule_output:
        return True

    # If it was initialized inside the model and nothing entered the module, it's bottom-level.
    if tensor_entry.initialized_inside_model and len(submodule.tl_tensors_entered_labels) == 0:
        tensor_entry.is_bottom_level_submodule_output = True
        tensor_entry.bottom_level_submodule_pass_exited = (
            submodule_address,
            submodule.tl_module_pass_num,
        )
        return True

    # Else, all parents must have entered the submodule for it to be a bottom-level submodule.
    for parent_label in tensor_entry.parent_layers:
        parent_tensor = model_log[parent_label]
        parent_modules_entered = parent_tensor.modules_entered
        if (len(parent_modules_entered) == 0) or (parent_modules_entered[-1] != submodule_address):
            tensor_entry.is_bottom_level_submodule_output = False
            return False

    # If it survived the above tests, it's a bottom-level submodule.
    tensor_entry.is_bottom_level_submodule_output = True
    tensor_entry.bottom_level_submodule_pass_exited = (
        submodule_address,
        submodule.tl_module_pass_num,
    )
    return True


def get_all_submodules(model: nn.Module, is_top_level_model: bool = True) -> List[nn.Module]:
    """Recursively gets list of all submodules for given module, no matter their level in the
    hierarchy; this includes the model itself.

    Args:
        model: PyTorch model.
        is_top_level_model: Whether it's the top-level model; just for the recursive logic of it.

    Returns:
        List of all submodules.
    """
    submodules = []
    if is_top_level_model:
        submodules.append(model)
    for module in model.children():
        if module not in submodules:
            submodules.append(module)
        submodules += get_all_submodules(module, is_top_level_model=False)
    return submodules


def cleanup_model(
    model_log,
    model: nn.Module,
    module_orig_forward_funcs: Dict[nn.Module, Callable],
    decorated_func_mapper: Dict[Callable, Callable],
):
    """Reverses all temporary changes to the model (namely, the forward hooks and added
    model attributes) that were added for PyTorch x-ray (scout's honor; leave no trace).

    Args:
        model: PyTorch model.
        module_orig_forward_funcs: Dict containing the original, undecorated forward pass functions for
            each submodule
        decorated_func_mapper: Dict mapping between original and decorated PyTorch funcs

    Returns:
        Original version of the model.
    """
    submodules = get_all_submodules(model, is_top_level_model=True)
    for submodule in submodules:
        if submodule == model:
            continue
        if submodule in module_orig_forward_funcs:
            submodule.forward = module_orig_forward_funcs[submodule]
    restore_model_attributes(
        model, decorated_func_mapper=decorated_func_mapper, attribute_keyword="tl"
    )
    undecorate_model_tensors(model)


def clear_hooks(hook_handles: List):
    """Takes in a list of tuples (module, hook_handle), and clears the hook at that
    handle for each module.

    Args:
        hook_handles: List of tuples (module, hook_handle)

    Returns:
        Nothing.
    """
    for hook_handle in hook_handles:
        hook_handle.remove()


def restore_module_attributes(
    module: nn.Module,
    decorated_func_mapper: Dict[Callable, Callable],
    attribute_keyword: str = "tl",
):
    def del_attrs_with_prefix(module, attribute_name):
        if attribute_name.startswith(attribute_keyword):
            delattr(module, attribute_name)
            return True

    for attribute_name, attr in iter_accessible_attributes(
        module, short_circuit=del_attrs_with_prefix
    ):
        if (
            isinstance(attr, Callable)
            and (attr in decorated_func_mapper)
            and (attribute_name[0:2] != "__")
        ):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                setattr(module, attribute_name, decorated_func_mapper[attr])


def restore_model_attributes(
    model: nn.Module,
    decorated_func_mapper: Dict[Callable, Callable],
    attribute_keyword: str = "tl",
):
    """Recursively clears the given attribute from all modules in the model.

    Args:
        model: PyTorch model.
        decorated_func_mapper: Dict mapping between original and decorated PyTorch funcs
        attribute_keyword: Any attribute with this keyword will be cleared.

    Returns:
        Nothing.
    """
    for module in get_all_submodules(model):
        restore_module_attributes(
            module,
            decorated_func_mapper=decorated_func_mapper,
            attribute_keyword=attribute_keyword,
        )

    for param in model.parameters():
        if hasattr(param, "tl_requires_grad"):
            param.requires_grad = getattr(param, "tl_requires_grad")
            delattr(param, "tl_requires_grad")


def undecorate_model_tensors(model: nn.Module):
    """Goes through a model and all its submodules, and unmutates any tensor attributes. Normally just clearing
    parameters would have done this, but some module types (e.g., batchnorm) contain attributes that are tensors,
    but not parameters.

    Args:
        model: PyTorch model

    Returns:
        PyTorch model with unmutated versions of all tensor attributes.
    """
    submodules = get_all_submodules(model)
    for submodule in submodules:
        for attribute_name, attribute in iter_accessible_attributes(submodule):
            if issubclass(type(attribute), torch.Tensor):
                if not issubclass(type(attribute), torch.nn.Parameter) and hasattr(
                    attribute, "tl_tensor_label_raw"
                ):
                    delattr(attribute, "tl_tensor_label_raw")
                    if hasattr(attribute, "tl_buffer_address"):
                        delattr(attribute, "tl_buffer_address")
                    if hasattr(attribute, "tl_buffer_parent"):
                        delattr(attribute, "tl_buffer_parent")
                else:
                    remove_attributes_starting_with_str(attribute, "tl_")
            elif type(attribute) in [list, tuple, set]:
                for item in attribute:
                    if issubclass(type(item), torch.Tensor) and hasattr(
                        item, "tl_tensor_label_raw"
                    ):
                        delattr(item, "tl_tensor_label_raw")
                        if hasattr(item, "tl_buffer_address"):
                            delattr(item, "tl_buffer_address")
                        if hasattr(item, "tl_buffer_parent"):
                            delattr(item, "tl_buffer_parent")
            elif type(attribute) == dict:
                for key, val in attribute.items():
                    if issubclass(type(val), torch.Tensor) and hasattr(val, "tl_tensor_label_raw"):
                        delattr(val, "tl_tensor_label_raw")
                        if hasattr(val, "tl_buffer_address"):
                            delattr(val, "tl_buffer_address")
                        if hasattr(val, "tl_buffer_parent"):
                            delattr(val, "tl_buffer_parent")



from collections import defaultdict
from typing import Any, Dict, List, Set, TYPE_CHECKING, Union

import torch

from .tensor_log import TensorLogEntry

if TYPE_CHECKING:
    pass

from .helper_funcs import (
    log_current_rng_states,
    set_rng_from_saved_states,
    tuple_tolerant_assign,
    tensor_nanequal,
    tensor_all_nan,
)

FUNCS_NOT_TO_PERTURB_IN_VALIDATION = [
    "expand_as",
    "new_zeros",
    "new_ones",
    "zero_",
    "copy_",
    "clamp",
    "fill_",
    "zeros_like",
    "ones_like",
]


def validate_saved_activations(
    self, ground_truth_output_tensors: List[torch.Tensor], verbose: bool = False
) -> bool:
    """Starting from outputs and internally terminated tensors, checks whether computing their values from the saved
    values of their input tensors yields their actually saved values, and whether computing their values from
    their parent tensors yields their saved values.

    Returns:
        True if it passes the tests, False otherwise.
    """
    # First check that the ground truth output tensors are accurate:
    for i, output_layer_label in enumerate(self.output_layers):
        output_layer = self[output_layer_label]
        if not tensor_nanequal(
            output_layer.tensor_contents,
            ground_truth_output_tensors[i],
            allow_tolerance=False,
        ):
            print(
                f"The {i}th output layer, {output_layer_label}, does not match the ground truth output tensor."
            )
            return False

    # Validate the parents of each validated layer.
    validated_child_edges_for_each_layer = defaultdict(set)
    validated_layers = set(self.output_layers + self.internally_terminated_layers)
    layers_to_validate_parents_for = list(validated_layers)

    while len(layers_to_validate_parents_for) > 0:
        layer_to_validate_parents_for = layers_to_validate_parents_for.pop(0)
        parent_layers_valid = validate_parents_of_saved_layer(
            self,
            layer_to_validate_parents_for,
            validated_layers,
            validated_child_edges_for_each_layer,
            layers_to_validate_parents_for,
            verbose,
        )
        if not parent_layers_valid:
            return False

    if len(validated_layers) < len(self.layer_labels):
        unreached = set(self.layer_labels) - validated_layers
        print(
            f"All saved activations were accurate, but some layers were not reached (check that "
            f"child args logged accurately): {unreached}"
        )
        return False

    return True


def validate_parents_of_saved_layer(
    self,
    layer_to_validate_parents_for_label: str,
    validated_layers: Set[str],
    validated_child_edges_for_each_layer: Dict[str, Set[str]],
    layers_to_validate_parents_for: List[str],
    verbose: bool = False,
) -> bool:
    """Given a layer, checks that 1) all parent tensors appear properly in the saved arguments for that layer,
    2) that executing the function for that layer with the saved parent layer activations yields the
    ground truth activation values for that layer, and 3) that plugging in "perturbed" values for each
    child layer yields values different from the saved activations for that layer.

    Args:
        layer_to_validate_parents_for_label:
        validated_layers:
        validated_child_edges_for_each_layer:
        layers_to_validate_parents_for:
        verbose: whether to print warning messages
    """
    layer_to_validate_parents_for = self[layer_to_validate_parents_for_label]

    # Check that the arguments are logged correctly:
    if not _check_layer_arguments_logged_correctly(self, layer_to_validate_parents_for_label):
        print(
            f"Parent arguments for layer {layer_to_validate_parents_for_label} are not logged properly; "
            f"either a parent wasn't logged as an argument, or was logged an extra time"
        )
        return False

    # Check that executing the function based on the actual saved values of the parents yields the saved
    # values of the layer itself:

    if not _check_whether_func_on_saved_parents_yields_saved_tensor(
        self, layer_to_validate_parents_for_label, perturb=False
    ):
        return False

    # Check that executing the layer's function on the wrong version of the saved parent tensors
    # yields the wrong tensors, when each saved tensor is perturbed in turn:

    for perturb_layer in layer_to_validate_parents_for.parent_layers:
        if layer_to_validate_parents_for.func_applied_name in FUNCS_NOT_TO_PERTURB_IN_VALIDATION:
            continue
        if not _check_whether_func_on_saved_parents_yields_saved_tensor(
            self,
            layer_to_validate_parents_for_label,
            perturb=True,
            layers_to_perturb=[perturb_layer],
            verbose=verbose,
        ):
            return False

    # Log that each parent layer has been validated for this source layer.

    for parent_layer_label in layer_to_validate_parents_for.parent_layers:
        parent_layer = self[parent_layer_label]
        validated_child_edges_for_each_layer[parent_layer_label].add(
            layer_to_validate_parents_for_label
        )
        if validated_child_edges_for_each_layer[parent_layer_label] == set(
            parent_layer.child_layers
        ):
            validated_layers.add(parent_layer_label)
            if (not parent_layer.is_input_layer) and not (
                parent_layer.is_buffer_layer and (parent_layer.buffer_parent is None)
            ):
                layers_to_validate_parents_for.append(parent_layer_label)

    return True


def _check_layer_arguments_logged_correctly(self, target_layer_label: str) -> bool:
    """Check whether the activations of the parent layers match the saved arguments of
    the target layer, and that the argument locations have been logged correctly.

    Args:
        target_layer_label: Layer to check

    Returns:
        True if arguments logged accurately, False otherwise
    """
    target_layer = self[target_layer_label]

    # Make sure that all parent layers appear in at least one argument and that no extra layers appear:
    parent_layers_in_args = set()
    for arg_type in ["args", "kwargs"]:
        parent_layers_in_args.update(list(target_layer.parent_layer_arg_locs[arg_type].values()))
    if parent_layers_in_args != set(target_layer.parent_layers):
        return False

    argtype_dict = {
        "args": (enumerate, "creation_args"),
        "kwargs": (lambda x: x.items(), "creation_kwargs"),
    }

    # Check for each parent layer that it is logged as a saved argument when it matches an argument, and
    # is not logged when it does not match a saved argument.

    for parent_layer_label in target_layer.parent_layers:
        parent_layer = self[parent_layer_label]
        for arg_type in ["args", "kwargs"]:
            iterfunc, argtype_field = argtype_dict[arg_type]
            for key, val in iterfunc(getattr(target_layer, argtype_field)):
                validation_correct_for_arg_and_layer = _validate_layer_against_arg(
                    self, target_layer, parent_layer, arg_type, key, val
                )
                if not validation_correct_for_arg_and_layer:
                    return False
    return True


def _validate_layer_against_arg(self, target_layer, parent_layer, arg_type, key, val):
    if type(val) in [list, tuple]:
        for v, subval in enumerate(val):
            argloc_key = (key, v)
            validation_correct_for_arg_and_layer = _check_arglocs_correct_for_arg(
                self, target_layer, parent_layer, arg_type, argloc_key, subval
            )
            if not validation_correct_for_arg_and_layer:
                return False

    elif type(val) == dict:
        for subkey, subval in val.items():
            argloc_key = (key, subkey)
            validation_correct_for_arg_and_layer = _check_arglocs_correct_for_arg(
                self, target_layer, parent_layer, arg_type, argloc_key, subval
            )
            if not validation_correct_for_arg_and_layer:
                return False
    else:
        argloc_key = key
        validation_correct_for_arg_and_layer = _check_arglocs_correct_for_arg(
            self, target_layer, parent_layer, arg_type, argloc_key, val
        )
        if not validation_correct_for_arg_and_layer:
            return False

    return True


def _check_arglocs_correct_for_arg(
    self,
    target_layer: TensorLogEntry,
    parent_layer: TensorLogEntry,
    arg_type: str,
    argloc_key: Union[str, tuple],
    saved_arg_val: Any,
):
    """For a given layer and an argument to its child layer, checks that it is logged correctly:
    that is, that it's logged as an argument if it matches, and is not logged as an argument if it doesn't match.
    """
    target_layer_label = target_layer.layer_label
    parent_layer_label = parent_layer.layer_label
    if target_layer_label in parent_layer.children_tensor_versions:
        parent_activations = parent_layer.children_tensor_versions[target_layer_label]
    else:
        parent_activations = parent_layer.tensor_contents

    if type(saved_arg_val) == torch.Tensor:
        parent_layer_matches_arg = tensor_nanequal(
            saved_arg_val, parent_activations, allow_tolerance=False
        )
    else:
        parent_layer_matches_arg = False
    parent_layer_logged_as_arg = (argloc_key in target_layer.parent_layer_arg_locs[arg_type]) and (
        target_layer.parent_layer_arg_locs[arg_type][argloc_key] == parent_layer_label
    )

    if (
        parent_layer_matches_arg
        and (not parent_layer_logged_as_arg)
        and (parent_activations.numel() != 0)
        and (parent_activations.dtype != torch.bool)
        and (not tensor_all_nan(parent_activations))
        and (parent_activations.abs().float().mean() != 0)
        and (parent_activations.abs().float().mean() != 1)
        and not any(
            [
                torch.equal(parent_activations, self[other_parent].tensor_contents)
                for other_parent in target_layer.parent_layers
                if other_parent != parent_layer_label
            ]
        )
    ):
        print(
            f"Parent {parent_layer_label} of {target_layer_label} has activations that match "
            f"{arg_type} {argloc_key} for {target_layer_label}, but is not logged as "
            f"such in parent_layer_arg_locs."
        )
        return False

    if (
        not parent_layer_matches_arg
        and parent_layer_logged_as_arg
        and parent_layer.func_applied_name in ["bernoulli_", "full"]
    ):  # in-place ops may modify tensor after logging, causing arg mismatch
        return True

    if (not parent_layer_matches_arg) and parent_layer_logged_as_arg:
        print(
            f"Parent {parent_layer_label} of {target_layer_label} is logged as {arg_type} {argloc_key} to "
            f"{target_layer_label}, but its saved activations don't match the saved argument."
        )
        return False

    return True


def _check_whether_func_on_saved_parents_yields_saved_tensor(
    self,
    layer_to_validate_parents_for_label: str,
    perturb: bool = False,
    layers_to_perturb: List[str] = None,
    verbose: bool = False,
) -> bool:
    """Checks whether executing the saved function for a layer on the saved value of its parent layers
    in fact yields the saved activations for that layer.

    Args:
        layer_to_validate_parents_for_label: label of the layer to check the saved activations
        perturb: whether to perturb the saved activations
        layers_to_perturb: layers for which to perturb the saved activations

    Returns:
        True if the activations match, False otherwise
    """
    if layers_to_perturb is None:
        layers_to_perturb = []

    layer_to_validate_parents_for = self[layer_to_validate_parents_for_label]

    if (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "__getitem__")
        and (type(layer_to_validate_parents_for.creation_args[1]) == torch.Tensor)
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[1],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "__getitem__")
        and not torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[0],
        )
    ):
        return True
    elif layer_to_validate_parents_for.func_applied_name == "empty_like":
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "__setitem__")
        and (type(layer_to_validate_parents_for.creation_args[1]) == torch.Tensor)
        and (layer_to_validate_parents_for.creation_args[1].dtype == torch.bool)
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[1],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "cross_entropy")
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[1],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "__setitem__")
        and (type(layer_to_validate_parents_for.creation_args[1]) == tuple)
        and (type(layer_to_validate_parents_for.creation_args[1][0]) == torch.Tensor)
        and (layer_to_validate_parents_for.creation_args[1][0].dtype == torch.bool)
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[1][0],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "index_select")
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[2],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "lstm")
        and (
            torch.equal(
                self[layers_to_perturb[0]].tensor_contents,
                layer_to_validate_parents_for.creation_args[1][0],
            )
            or torch.equal(
                self[layers_to_perturb[0]].tensor_contents,
                layer_to_validate_parents_for.creation_args[1][1],
            )
            or torch.equal(
                self[layers_to_perturb[0]].tensor_contents,
                layer_to_validate_parents_for.creation_args[2][0],
            )
            or torch.equal(
                self[layers_to_perturb[0]].tensor_contents,
                layer_to_validate_parents_for.creation_args[2][1],
            )
            or (
                (type(layer_to_validate_parents_for.creation_args[1]) == torch.Tensor)
                and torch.equal(
                    self[layers_to_perturb[0]].tensor_contents,
                    layer_to_validate_parents_for.creation_args[1],
                )
            )
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "_pad_packed_sequence")
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[1],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "masked_fill_")
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[1],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "scatter_")
        and torch.equal(
            self[layers_to_perturb[0]].tensor_contents,
            layer_to_validate_parents_for.creation_args[2],
        )
    ):
        return True
    elif (
        perturb
        and (layer_to_validate_parents_for.func_applied_name == "interpolate")
        and (
            (
                ("scale_factor" in layer_to_validate_parents_for.creation_kwargs)
                and (layer_to_validate_parents_for.creation_kwargs["scale_factor"] is not None)
                and torch.equal(
                    self[layers_to_perturb[0]].tensor_contents,
                    torch.tensor(layer_to_validate_parents_for.creation_kwargs["scale_factor"]),
                )
            )
            or (
                (len(layer_to_validate_parents_for.creation_args) >= 3)
                and torch.equal(
                    self[layers_to_perturb[0]].tensor_contents,
                    layer_to_validate_parents_for.creation_args[2],
                )
            )
        )
    ):
        return True

    # Prepare input arguments: keep the ones that should just be kept, perturb those that should be perturbed

    input_args = _prepare_input_args_for_validating_layer(
        self, layer_to_validate_parents_for, layers_to_perturb
    )

    # set the saved rng value:
    layer_func = layer_to_validate_parents_for.func_applied
    current_rng_states = log_current_rng_states()
    set_rng_from_saved_states(layer_to_validate_parents_for.func_rng_states)
    try:
        recomputed_output = layer_func(*input_args["args"], **input_args["kwargs"])
    except Exception:
        # Perturbed arguments are invalid for this function (e.g., pack_padded_sequence
        # requires valid lengths). Treat as a valid exemption.
        set_rng_from_saved_states(current_rng_states)
        return True
    set_rng_from_saved_states(current_rng_states)

    if layer_func.__name__ in [
        "__setitem__",
        "zero_",
        "__delitem__",
    ]:  # TODO: fix this
        recomputed_output = input_args["args"][0]

    if any([issubclass(type(recomputed_output), which_type) for which_type in [list, tuple]]):
        recomputed_output = recomputed_output[layer_to_validate_parents_for.iterable_output_index]

    if (
        not (
            tensor_nanequal(
                recomputed_output,
                layer_to_validate_parents_for.tensor_contents,
                allow_tolerance=True,
            )
        )
        and not perturb
    ):
        # In-place RNG ops (e.g. bernoulli_) modify tensor after logging, so
        # parents' saved values are stale when recomputing child output.
        parent_has_inplace_rng = any(
            self[p].func_applied_name in ["bernoulli_", "full"]
            for p in layer_to_validate_parents_for.parent_layers
        )
        if parent_has_inplace_rng:
            return True
        print(
            f"Saved activations for layer {layer_to_validate_parents_for_label} do not match the "
            f"values computed based on the parent layers {layer_to_validate_parents_for.parent_layers}."
        )
        return False

    if (
        tensor_nanequal(
            recomputed_output,
            layer_to_validate_parents_for.tensor_contents,
            allow_tolerance=False,
        )
        and perturb
    ):
        return _posthoc_perturb_check(
            self, layer_to_validate_parents_for, layers_to_perturb, verbose
        )

    return True


def _prepare_input_args_for_validating_layer(
    self,
    layer_to_validate_parents_for: TensorLogEntry,
    layers_to_perturb: List[str],
) -> Dict:
    """Prepares the input arguments for validating the saved activations of a layer.

    Args:
        layer_to_validate_parents_for: Layer being checked.
        layers_to_perturb: Layers for which to perturb the saved activations.

    Returns:
        Dict of input arguments.
    """
    input_args = {
        "args": list(layer_to_validate_parents_for.creation_args[:]),
        "kwargs": layer_to_validate_parents_for.creation_kwargs.copy(),
    }
    input_args = _copy_validation_args(input_args)

    # Swap in saved parent activations:

    for arg_type in ["args", "kwargs"]:
        for (
            key,
            parent_layer_arg,
        ) in layer_to_validate_parents_for.parent_layer_arg_locs[arg_type].items():
            parent_layer = self[parent_layer_arg]
            if layer_to_validate_parents_for.layer_label in parent_layer.children_tensor_versions:
                parent_values = parent_layer.children_tensor_versions[
                    layer_to_validate_parents_for.layer_label
                ]
            else:
                parent_values = parent_layer.tensor_contents
            parent_values = parent_values.detach().clone()

            if parent_layer_arg in layers_to_perturb:
                parent_layer_func_values = _perturb_layer_activations(
                    parent_values, layer_to_validate_parents_for.tensor_contents
                )
            else:
                parent_layer_func_values = parent_values

            if type(key) != tuple:
                input_args[arg_type][key] = parent_layer_func_values
            else:
                input_args[arg_type][key[0]] = tuple_tolerant_assign(
                    input_args[arg_type][key[0]], key[1], parent_layer_func_values
                )

    return input_args


def _copy_validation_args(input_args: Dict):
    new_args = []
    for i, val in enumerate(input_args["args"]):
        if type(val) == torch.Tensor:
            new_args.append(val.detach().clone())
        elif type(val) in [list, tuple, set]:
            new_iter = []
            for i2, val2 in enumerate(val):
                if type(val2) == torch.Tensor:
                    new_iter.append(val2.detach().clone())
                else:
                    new_iter.append(val2)
            new_args.append(type(val)(new_iter))
        else:
            new_args.append(val)
    input_args["args"] = new_args

    new_kwargs = {}
    for key, val in input_args["kwargs"].items():
        if type(val) == torch.Tensor:
            new_kwargs[key] = val.detach().clone()
        elif type(val) in [list, tuple, set]:
            new_iter = []
            for i2, val2 in enumerate(val):
                if type(val2) == torch.Tensor:
                    new_iter.append(val2.detach().clone())
                else:
                    new_iter.append(val2)
            new_kwargs[key] = type(val)(new_iter)
        else:
            new_kwargs[key] = val
    input_args["kwargs"] = new_kwargs
    return input_args


def _perturb_layer_activations(
    parent_activations: torch.Tensor, output_activations: torch.Tensor
) -> torch.Tensor:
    """
    Perturbs the values of a saved tensor.

    Args:
        parent_activations: Tensor of activation values for the parent tensor
        output_activations: Tensor of activation values for the tensor whose parents are being tested (the output)

    Returns:
        Perturbed version of saved tensor
    """
    device = parent_activations.device
    if parent_activations.numel() == 0:
        return parent_activations.detach().clone()

    if parent_activations.dtype in [
        torch.int,
        torch.long,
        torch.short,
        torch.uint8,
        torch.int8,
        torch.int16,
        torch.int32,
        torch.int64,
    ]:
        tensor_unique_vals = torch.unique(parent_activations)
        if len(tensor_unique_vals) > 1:
            perturbed_activations = parent_activations.detach().clone()
            while torch.equal(perturbed_activations, parent_activations):
                perturbed_activations = torch.randint(
                    parent_activations.min(),
                    parent_activations.max() + 1,
                    size=parent_activations.shape,
                    device=device,
                ).type(parent_activations.dtype)
        else:
            perturbed_activations = parent_activations.detach().clone()
            while torch.equal(perturbed_activations, parent_activations):
                if torch.min(parent_activations) < 0:
                    perturbed_activations = torch.randint(
                        -10, 11, size=parent_activations.shape, device=device
                    ).type(parent_activations.dtype)
                else:
                    perturbed_activations = torch.randint(
                        0, 11, size=parent_activations.shape, device=device
                    ).type(parent_activations.dtype)

    elif parent_activations.dtype == torch.bool:
        perturbed_activations = parent_activations.detach().clone()
        while torch.equal(perturbed_activations, parent_activations):
            perturbed_activations = torch.randint(
                0, 2, size=parent_activations.shape, device=device
            ).bool()
    else:
        mean_output = output_activations.detach().float().abs().mean()
        mean_output += torch.rand(mean_output.shape, device=mean_output.device) * 100
        mean_output *= torch.rand(mean_output.shape, device=mean_output.device)
        mean_output.requires_grad = False
        scale = mean_output.to(device)
        if parent_activations.is_complex():
            perturbed_activations = torch.complex(
                torch.randn(parent_activations.shape, device=device) * scale,
                torch.randn(parent_activations.shape, device=device) * scale,
            ).type(parent_activations.dtype)
        else:
            perturbed_activations = (
                torch.randn_like(parent_activations.float(), device=device) * scale
            )
            perturbed_activations = perturbed_activations.type(parent_activations.dtype)

    return perturbed_activations


def _posthoc_perturb_check(
    self,
    layer_to_validate_parents_for: TensorLogEntry,
    layers_to_perturb: List[str],
    verbose: bool = False,
) -> bool:
    """If a layer fails the "perturbation check"--that is, if perturbing the values of parent
    layers doesn't change the values relative to the layer's saved values--checks whether one of the
    remaining arguments is a "special" tensor, such as all-ones or all-zeros, such that perturbing a tensor
    wouldn't necessarily change the output of the layer.

    Args:
        layer_to_validate_parents_for: layer being checked.
        layers_to_perturb: parent layers being perturbed

    Returns:
        True if there's an "excuse" for the perturbation failing, False otherwise.
    """
    # Check if the tensor is all nans or all infinite:
    if layer_to_validate_parents_for.tensor_dtype == torch.bool:
        return True
    elif (
        (layer_to_validate_parents_for.func_applied_name == "to")
        and (len(layer_to_validate_parents_for.creation_args) > 1)
        and (type(layer_to_validate_parents_for.creation_args[1]) == torch.Tensor)
    ):
        return True
    elif (
        (layer_to_validate_parents_for.func_applied_name == "__setitem__")
        and (type(layer_to_validate_parents_for.creation_args[2]) == torch.Tensor)
        and (
            layer_to_validate_parents_for.creation_args[0].shape
            == layer_to_validate_parents_for.creation_args[2].shape
        )
    ):
        return True
    elif (layer_to_validate_parents_for.func_applied_name == "__setitem__") and not isinstance(
        layer_to_validate_parents_for.creation_args[2], torch.Tensor
    ):
        return True
    elif (layer_to_validate_parents_for.func_applied_name in ["__getitem__", "unbind"]) and (
        layer_to_validate_parents_for.tensor_contents.numel() < 20
    ):  # some elements can be the same by chance
        return True
    elif layer_to_validate_parents_for.func_applied_name in [
        "meshgrid",
        "broadcast_tensors",
    ]:  # output[i] depends only on input[i], but all inputs logged as parents of all outputs
        return True
    elif layer_to_validate_parents_for.func_applied_name in [
        "full_like",
        "zeros_like",
        "ones_like",
        "empty_like",
        "rand_like",
        "randn_like",
    ]:  # *_like ops use input only for shape/dtype/device, not values
        return True
    elif (
        (layer_to_validate_parents_for.func_applied_name == "__getitem__")
        and (type(layer_to_validate_parents_for.creation_args[1]) == torch.Tensor)
        and (len(layer_to_validate_parents_for.creation_args[1].unique()) < 20)
    ):
        return True
    elif (layer_to_validate_parents_for.func_applied_name == "max") and len(
        layer_to_validate_parents_for.creation_args
    ) > 1:
        return True
    elif (layer_to_validate_parents_for.func_applied_name == "max") and not torch.is_floating_point(
        layer_to_validate_parents_for.creation_args[0]
    ):
        return True
    else:
        num_inf = torch.isinf(layer_to_validate_parents_for.tensor_contents.abs()).int().sum()
        num_nan = torch.isnan(layer_to_validate_parents_for.tensor_contents.abs()).int().sum()
        if (num_inf == layer_to_validate_parents_for.tensor_contents.numel()) or (
            num_nan == layer_to_validate_parents_for.tensor_contents.numel()
        ):
            return True

    arg_type_dict = {
        "args": (enumerate, "creation_args"),
        "kwargs": (lambda x: x.items(), "creation_kwargs"),
    }

    layer_to_validate_parents_for_label = layer_to_validate_parents_for.layer_label
    for arg_type in ["args", "kwargs"]:
        iterfunc, fieldname = arg_type_dict[arg_type]
        for key, val in iterfunc(getattr(layer_to_validate_parents_for, fieldname)):
            # Skip if it's the argument itself:
            if (key in layer_to_validate_parents_for.parent_layer_arg_locs[arg_type]) and (
                layer_to_validate_parents_for.parent_layer_arg_locs[arg_type][key]
            ) in layers_to_perturb:
                continue
            arg_is_special = _check_if_arg_is_special_val(val)
            if arg_is_special:
                if verbose:
                    print(
                        f"Activations for layer {layer_to_validate_parents_for_label} do not change when "
                        f"values for {layers_to_perturb} are changed (out of parent "
                        f"layers {layer_to_validate_parents_for.parent_layers}), but {arg_type[:-1]} {key} is "
                        f"all zeros or all-ones, so validation still succeeds..."
                    )
                return True

    print(
        f"Activations for layer {layer_to_validate_parents_for_label} do not change when "
        f"values for {layers_to_perturb} are changed (out of parent "
        f"layers {layer_to_validate_parents_for.parent_layers}), and the other "
        f'arguments are not "special" (all-ones or all-zeros) tensors.'
    )
    return False


def _check_if_arg_is_special_val(val: Union[torch.Tensor, Any]):
    # If it's one of the other arguments, check if it's all zeros or all ones:
    if type(val) != torch.Tensor:
        try:
            val = torch.tensor(val)
        except (TypeError, ValueError, RuntimeError):
            return False
    if torch.all(torch.eq(val, 0)) or torch.all(torch.eq(val, 1)) or (val.numel() == 0):
        return True
    else:
        return False



import warnings

import torch

from .constants import MODEL_HISTORY_FIELD_ORDER
from .helper_funcs import remove_entry_from_list
from .tensor_log import TensorLogEntry


def cleanup(self):
    """Deletes all log entries in the model."""
    for tensor_log_entry in self:
        self._remove_log_entry(tensor_log_entry, remove_references=True)
    for attr in MODEL_HISTORY_FIELD_ORDER:
        delattr(self, attr)
    torch.cuda.empty_cache()


def _remove_log_entry(self, log_entry: TensorLogEntry, remove_references: bool = True):
    """Given a TensorLogEntry, destroys it and all references to it.

    Args:
        log_entry: Tensor log entry to remove.
        remove_references: Whether to also remove references to the log entry
    """
    if self._pass_finished:
        tensor_label = log_entry.layer_label
    else:
        tensor_label = log_entry.tensor_label_raw
    for attr in dir(log_entry):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if not attr.startswith("__") and not callable(getattr(log_entry, attr)):
                delattr(log_entry, attr)
    del log_entry
    if remove_references:
        _remove_log_entry_references(self, tensor_label)


def _remove_log_entry_references(self, layer_to_remove: str):
    """Removes all references to a given TensorLogEntry in the ModelHistory object.

    Args:
        layer_to_remove: The log entry to remove.
    """
    # Clear any fields in ModelHistory referring to the entry.

    remove_entry_from_list(self.input_layers, layer_to_remove)
    remove_entry_from_list(self.output_layers, layer_to_remove)
    remove_entry_from_list(self.buffer_layers, layer_to_remove)
    remove_entry_from_list(self.internally_initialized_layers, layer_to_remove)
    remove_entry_from_list(self.internally_terminated_layers, layer_to_remove)
    remove_entry_from_list(self.internally_terminated_bool_layers, layer_to_remove)
    remove_entry_from_list(self.layers_with_saved_activations, layer_to_remove)
    remove_entry_from_list(self.layers_with_saved_gradients, layer_to_remove)
    remove_entry_from_list(self._layers_where_internal_branches_merge_with_input, layer_to_remove)

    self.conditional_branch_edges = [
        tup for tup in self.conditional_branch_edges if layer_to_remove not in tup
    ]

    # Now any nested fields.

    for group_label, group_tensors in self.layers_computed_with_params.items():
        if layer_to_remove in group_tensors:
            group_tensors.remove(layer_to_remove)
    self.layers_computed_with_params = {
        k: v for k, v in self.layers_computed_with_params.items() if len(v) > 0
    }

    for group_label, group_tensors in self.equivalent_operations.items():
        if layer_to_remove in group_tensors:
            group_tensors.remove(layer_to_remove)
    self.equivalent_operations = {k: v for k, v in self.equivalent_operations.items() if len(v) > 0}



import inspect
import time
import types
import warnings
from functools import wraps
from typing import Callable, Dict, List, TYPE_CHECKING, Tuple

import torch

from .constants import ORIG_TORCH_FUNCS
from .helper_funcs import (
    get_vars_of_type_from_obj,
    identity,
    log_current_rng_states,
    make_random_barcode,
    nested_getattr,
    print_override,
    safe_copy,
)
from .logging_funcs import log_function_output_tensors, log_source_tensor

if TYPE_CHECKING:
    from model_history import ModelHistory

funcs_not_to_log = ["numpy", "__array__", "size", "dim"]
print_funcs = ["__repr__", "__str__", "_str"]


def torch_func_decorator(self, func: Callable, func_name: str):
    @wraps(func)
    def wrapped_func(*args, **kwargs):
        # Initial bookkeeping; check if it's a special function, organize the arguments.
        self.current_function_call_barcode = 0
        if (func_name in funcs_not_to_log) or (not self._track_tensors) or self._pause_logging:
            out = func(*args, **kwargs)
            return out
        all_args = list(args) + list(kwargs.values())
        arg_tensorlike = get_vars_of_type_from_obj(all_args, torch.Tensor)

        # Register any buffer tensors in the arguments.

        for t in arg_tensorlike:
            if hasattr(t, "tl_buffer_address"):
                log_source_tensor(self, t, "buffer", getattr(t, "tl_buffer_address"))

        if (func_name in print_funcs) and (len(arg_tensorlike) > 0):
            out = print_override(args[0], func_name)
            return out

        # Copy the args and kwargs in case they change in-place:
        if self.save_function_args:
            arg_copies = tuple([safe_copy(arg) for arg in args])
            kwarg_copies = {k: safe_copy(v) for k, v in kwargs.items()}
        else:
            arg_copies = args
            kwarg_copies = kwargs

        # Call the function, tracking the timing, rng states, and whether it's a nested function
        func_call_barcode = make_random_barcode()
        self.current_function_call_barcode = func_call_barcode
        start_time = time.time()
        func_rng_states = log_current_rng_states()
        out_orig = func(*args, **kwargs)
        func_time_elapsed = time.time() - start_time
        is_bottom_level_func = self.current_function_call_barcode == func_call_barcode

        if func_name in ["__setitem__", "zero_", "__delitem__"]:
            out_orig = args[0]

        was_inplace = len(args) > 0 and id(out_orig) == id(args[0])
        if was_inplace:
            out_orig = safe_copy(out_orig)

        # Log all output tensors
        output_tensors = get_vars_of_type_from_obj(
            out_orig,
            which_type=torch.Tensor,
            subclass_exceptions=[torch.nn.Parameter],
        )

        if len(output_tensors) > 0:
            log_function_output_tensors(
                self,
                func,
                func_name,
                args,
                kwargs,
                arg_copies,
                kwarg_copies,
                out_orig,
                func_time_elapsed,
                func_rng_states,
                is_bottom_level_func,
            )

            # Propagate label to original tensor for in-place ops so
            # subsequent operations see the updated label.
            if was_inplace:
                if hasattr(out_orig, "tl_tensor_label_raw"):
                    args[0].tl_tensor_label_raw = out_orig.tl_tensor_label_raw

        return out_orig

    return wrapped_func


def decorate_pytorch(
    self: "ModelHistory", torch_module: types.ModuleType, orig_func_defs: List[Tuple]
) -> Dict[Callable, Callable]:
    """Mutates all PyTorch functions (TEMPORARILY!) to save the outputs of any functions
    that return Tensors, along with marking them with metadata. Returns a list of tuples that
    save the current state of the functions, such that they can be restored when done.

    args:
        torch_module: The top-level torch module (i.e., from "import torch").
            This is supplied as an argument on the off-chance that the user has imported torch
            and done their own monkey-patching.
        tensors_to_mutate: A list of tensors that will be mutated (since any tensors created
            before calling the torch mutation function will not be mutated).
        orig_func_defs: Supply a list from outside to guarantee it can be cleaned up properly.
        tensor_record: A list to which the outputs of the functions will be appended.

    returns:
        List of tuples consisting of [namespace, func_name, orig_func], sufficient
        to return torch to normal when finished, and also a dict mapping mutated functions to original functions.
    """

    # Do a pass to save the original func defs.
    collect_orig_func_defs(torch_module, orig_func_defs)
    decorated_func_mapper = {}

    # Get references to the function classes.
    function_class = type(lambda: 0)
    builtin_class = type(torch.mean)
    method_class = type(torch.Tensor.__add__)
    wrapper_class = type(torch.Tensor.__getitem__)
    getset_class = type(torch.Tensor.real)

    for namespace_name, func_name in ORIG_TORCH_FUNCS:
        namespace_name_notorch = namespace_name.replace("torch.", "")
        local_func_namespace = nested_getattr(torch_module, namespace_name_notorch)
        if not hasattr(local_func_namespace, func_name):
            continue
        orig_func = getattr(local_func_namespace, func_name)
        if func_name not in self.func_argnames:
            get_func_argnames(self, orig_func, func_name)
        if getattr(orig_func, "__name__", False) == "wrapped_func":
            continue

        if type(orig_func) in [function_class, builtin_class, method_class, wrapper_class]:
            new_func = torch_func_decorator(self, orig_func, func_name)
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    setattr(local_func_namespace, func_name, new_func)
                new_func.tl_is_decorated_function = True
                decorated_func_mapper[new_func] = orig_func
                decorated_func_mapper[orig_func] = new_func
            except (AttributeError, TypeError):
                pass

        elif type(orig_func) is getset_class:
            getter_orig, setter_orig, deleter_orig = (
                orig_func.__get__,
                orig_func.__set__,
                orig_func.__delete__,
            )
            getter_dec, setter_dec, deleter_dec = (
                torch_func_decorator(self, getter_orig, func_name),
                torch_func_decorator(self, setter_orig, func_name),
                torch_func_decorator(self, deleter_orig, func_name),
            )
            getter_dec.tl_is_decorated_function = True
            setter_dec.tl_is_decorated_function = True
            deleter_dec.tl_is_decorated_function = True
            new_property = property(getter_dec, setter_dec, deleter_dec, doc=func_name)
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    setattr(local_func_namespace, func_name, new_property)
            except (AttributeError, TypeError) as _:
                pass
            decorated_func_mapper[new_property] = orig_func
            decorated_func_mapper[orig_func] = new_property

    # Bolt on the identity function
    new_identity = torch_func_decorator(self, identity, "identity")
    torch.identity = new_identity

    return decorated_func_mapper


def undecorate_pytorch(
    torch_module, orig_func_defs: List[Tuple], input_tensors: List[torch.Tensor]
):
    """
    Returns all PyTorch functions back to the definitions they had when mutate_pytorch was called.
    This is done for the output tensors and history_dict too to avoid ugliness. Also deletes
    the mutant versions of the functions to remove any references to old ModelHistory object.

    args:
        torch_module: The torch module object.
        orig_func_defs: List of tuples consisting of [namespace_name, func_name, orig_func], sufficient
            to regenerate the original functions.
        input_tensors: List of input tensors whose fucntions will be undecorated.
        decorated_func_mapper: Maps the decorated function to the original function
    """
    for namespace_name, func_name, orig_func in orig_func_defs:
        namespace_name_notorch = namespace_name.replace("torch.", "")
        local_func_namespace = nested_getattr(torch_module, namespace_name_notorch)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            decorated_func = getattr(local_func_namespace, func_name)
        del decorated_func
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                setattr(local_func_namespace, func_name, orig_func)
        except (AttributeError, TypeError) as _:
            continue
    if hasattr(torch, "identity"):
        delattr(torch, "identity")
    for input_tensor in input_tensors:
        if hasattr(input_tensor, "tl_tensor_label_raw"):
            delattr(input_tensor, "tl_tensor_label_raw")


def collect_orig_func_defs(torch_module: types.ModuleType, orig_func_defs: List[Tuple]):
    """Collects the original torch function definitions, so they can be restored after the logging is done.

    Args:
        torch_module: The top-level torch module
        orig_func_defs: List of tuples keeping track of the original function definitions
    """
    for namespace_name, func_name in ORIG_TORCH_FUNCS:
        namespace_name_notorch = namespace_name.replace("torch.", "")
        local_func_namespace = nested_getattr(torch_module, namespace_name_notorch)
        if not hasattr(local_func_namespace, func_name):
            continue
        orig_func = getattr(local_func_namespace, func_name)
        orig_func_defs.append((namespace_name, func_name, orig_func))


# TODO: hard-code some of the arg names; for example truediv, getitem, etc. Can crawl through and see what isn't working
def get_func_argnames(self, orig_func: Callable, func_name: str):
    """Attempts to get the argument names for a function, first by checking the signature, then
    by checking the documentation. Adds these names to func_argnames if it can find them,
    doesn't do anything if it can't."""
    if func_name in ["real", "imag", "T", "mT", "data", "H"]:
        return

    try:
        argnames = list(inspect.signature(orig_func).parameters.keys())
        argnames = tuple([arg.replace("*", "") for arg in argnames if arg not in ["cls", "self"]])
        self.func_argnames[func_name] = argnames
        return
    except ValueError:
        pass

    docstring = orig_func.__doc__
    if (type(docstring) is not str) or (len(docstring) == 0):  # if docstring missing, skip it
        return

    open_ind, close_ind = docstring.find("("), docstring.find(")")
    argstring = docstring[open_ind + 1 : close_ind]
    arg_list = argstring.split(",")
    arg_list = [arg.strip(" ") for arg in arg_list]
    argnames = []
    for arg in arg_list:
        argname = arg.split("=")[0]
        if argname in ["*", "/", "//", ""]:
            continue
        argname = argname.replace("*", "")
        argnames.append(argname)
    argnames = tuple([arg for arg in argnames if arg not in ["self", "cls"]])
    self.func_argnames[func_name] = argnames
    return
